#!/usr/bin/env python3

import json
import logging
import sys
import warnings
from functools import wraps

import click
from ocviapy import apply_config, get_current_namespace, StatusError
from tabulate import tabulate
from wait_for import TimedOutError

import bonfire.config as conf
from bonfire.elastic_logging import ElasticLogger
from bonfire.local import get_local_apps, get_appsfile_apps
from bonfire.utils import AppOrComponentSelector, RepoFile, SYNTAX_ERR
from bonfire.namespaces import (
    Namespace,
    extend_namespace,
    get_namespaces,
    release_reservation,
    reserve_namespace,
    describe_namespace,
)
from bonfire.openshift import (
    check_for_existing_reservation,
    find_clowd_env_for_ns,
    get_namespace_pools,
    get_reservation,
    has_clowder,
    has_ns_operator,
    wait_for_all_resources,
    wait_for_clowd_env_target_ns,
    wait_for_db_resources,
    wait_on_cji,
    whoami,
    get_console_url,
    get_pool_size_limit,
    get_reserved_namespace_quantity,
)
from bonfire.processor import TemplateProcessor, process_clowd_env, process_iqe_cji
from bonfire.qontract import get_apps_for_env, sub_refs
from bonfire.secrets import import_secrets_from_dir
from bonfire.configmaps import import_configmaps_from_dir
from bonfire.utils import (
    FatalError,
    check_pypi,
    find_what_depends_on,
    get_version,
    split_equals,
    validate_time_string,
    merge_app_configs,
)


log = logging.getLogger(__name__)
es_telemetry = ElasticLogger()

APP_SRE_SRC = "appsre"
FILE_SRC = "file"
NO_RESERVATION_SYS = "this cluster does not use a namespace reservation system"

_local_option = click.option(
    "--local",
    help="Whether 'oc process' uses --local=true or --local=false (default: true)",
    type=bool,
    default=True,
)


def _error(msg):
    es_telemetry.send_telemetry(msg, success=False)
    click.echo(f"ERROR: {msg}", err=True)
    sys.exit(1)


def current_namespace_or_error():
    log.info("attempting to use current namespace from oc/kubectl context...")
    namespace = get_current_namespace()
    if not namespace:
        _error(
            "Namespace from current oc/kubectl context is not set."
            " Please specify namespace using options/arguments."
        )
    return namespace


def click_exception_wrapper(command):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            try:
                result = f(*args, **kwargs)
                return result
            except KeyboardInterrupt:
                _error(f"{command}: aborted by keyboard interrupt")
            except TimedOutError as err:
                _error(f"{command}: hit timeout error: {err}")
            except FatalError as err:
                _error(f"{command}: hit fatal error: {err}")
            except StatusError as err:
                _error(f"{command}: hit status error: {err}")
            except Exception as err:
                log.exception("hit unexpected error")
                _error(f"{command}: hit unexpected error: {err}")

        return wrapper

    return decorator


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option("--debug", "-d", help="Enable debug logging", is_flag=True, default=False)
def main(debug):
    logging.getLogger("sh").setLevel(logging.CRITICAL)  # silence the 'sh' library logger
    logging.basicConfig(
        format="%(asctime)s [%(levelname)8s] [%(threadName)20s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=logging.DEBUG if debug else logging.INFO,
    )

    def custom_formatwarning(msg, *args, **kwargs):
        # ignore everything except the message
        return str(msg)

    warnings.formatwarning = custom_formatwarning
    warnings.simplefilter("default")
    logging.captureWarnings(True)
    if conf.ENV_FILE:
        log.debug("using env file: %s", conf.ENV_FILE)

    check_pypi()


@main.group(hidden=True)
def test():
    """
    Used for unit testing
    """
    pass


@main.group()
def namespace():
    """Perform operations related to namespace reservation"""
    pass


@main.group()
def config():
    """Commands related to bonfire configuration"""
    pass


@main.group()
def apps():
    """Show information about deployable apps"""
    pass


@main.group()
def pool():
    """Perform operations related to pool types"""
    pass


def _confirm_or_abort(msg):
    if conf.BONFIRE_BOT:
        # these types of warnings shouldn't occur in automated runs, error out immediately
        _error(msg)
    else:
        # have end user confirm if they want to proceed anyway
        click.echo(f"\n{msg}")
        prompt = "Continue anyway?"
        if not sys.stdout.isatty():
            _error(
                f"Prompt cannot be answered:\n\n{msg}\n{prompt}\n\nOutput is not a TTY. Aborting."
            )
        if not click.confirm(prompt):
            click.echo("Aborting")
            sys.exit(0)


def _warn_if_not_owned_by_me():
    _confirm_or_abort("Namespace currently reserved by someone else")


def _warn_if_not_ready():
    _confirm_or_abort("Namespace's environment is not currently ready")


def _warn_before_delete():
    _confirm_or_abort("Deleting this reservation will also delete the associated namespace")


def _warn_of_existing(requester):
    _confirm_or_abort(
        f"Existing reservation(s) found for requester '{requester}', "
        "consider re-using the existing namespace."
    )


def _get_requester():
    if conf.BONFIRE_NS_REQUESTER:
        requester = conf.BONFIRE_NS_REQUESTER
    else:
        try:
            requester = whoami()
        except Exception:
            log.info("whoami returned an error - setting requester to 'bonfire'")  # minikube
            requester = "bonfire"
    return requester


def _wait_on_namespace_resources(namespace, timeout, db_only=False):
    if db_only:
        wait_for_db_resources(namespace, timeout)
    else:
        wait_for_all_resources(namespace, timeout)


def _validate_reservation_duration(ctx, param, value):
    try:
        return validate_time_string(value)
    except ValueError as err:
        raise click.BadParameter(err)


_ns_reserve_options = [
    click.option(
        "--name",
        type=str,
        default=None,
        help="Identifier for the reservation",
    ),
    click.option(
        "--requester",
        "-r",
        type=str,
        default=None,
        help="Name of the user requesting a reservation",
    ),
    click.option(
        "--duration",
        "-d",
        type=str,
        default="1h",
        help="Duration of the reservation",
        callback=_validate_reservation_duration,
    ),
    click.option(
        "--pool",
        type=str,
        default=conf.DEFAULT_NAMESPACE_POOL,
        show_default=True,
        help="Specifies the pool type name",
    ),
    click.option(
        "-f",
        "--force",
        is_flag=True,
        default=False,
        help="Don't prompt if reservations exist for user",
    ),
    _local_option,
]


_ns_list_options = [
    click.option(
        "--available",
        "-a",
        is_flag=True,
        default=False,
        help="show only un-reserved/ready namespaces",
    ),
    click.option(
        "--mine",
        "-m",
        is_flag=True,
        default=False,
        help="show only namespaces reserved in your name",
    ),
    click.option(
        "--output",
        "-o",
        default="cli",
        help="which output format to return the data in",
        type=click.Choice(["cli", "json"], case_sensitive=False),
    ),
]

_timeout_option = [
    click.option(
        "--timeout",
        "-t",
        required=True,
        type=int,
        default=600,
        help="timeout in sec (default = 600) to wait for resources to be ready",
    )
]


def _validate_set_template_ref(ctx, param, value):
    try:
        split_value = split_equals(value)
        if split_value:
            # check that values unpack properly
            for app_component, value in split_value.items():
                # TODO: remove once app name syntax fully deprecated
                split = app_component.split("/")
                if len(split) == 2:
                    warnings.warn(
                        (
                            "--set-template-ref: <app>/<component>=<ref> syntax is deprecated, "
                            "use <component>=<ref>"
                        ),
                        DeprecationWarning,
                    )
                elif len(split) > 2:
                    raise ValueError
        return split_value
    except ValueError:
        raise click.BadParameter("format must be '<component>=<ref>'")


def _validate_set_parameter(ctx, param, value):
    try:
        split_value = split_equals(value)
        if split_value:
            # check that values unpack properly
            for param_path, value in split_value.items():
                # TODO: remove once app name syntax fully deprecated
                split = param_path.split("/")
                if len(split) == 3:
                    warnings.warn(
                        (
                            "--set-parameter: <app>/<component>/<param>=<value> syntax is "
                            "deprecated, use <component>/<param>=<value>"
                        ),
                        DeprecationWarning,
                    )
                elif len(split) < 2 or len(split) > 3:
                    raise ValueError
        return split_value
    except ValueError:
        raise click.BadParameter("format must be '<component>/<param>=<value>'")


def _validate_split_equals(ctx, param, value):
    try:
        return split_equals(value)
    except ValueError:
        msg = "format must be '<name>=<value>'"
        name = param.name
        if name == "set_image_tag":
            msg = "format must be '<image uri>=<tag>', example: 'quay.io/org/repo=latest'"
        elif name == "preferred_params":
            msg = "format must be 'PARAMETER_NAME=value'"
        elif name == "custom_env_vars":
            msg = "format must be 'ENV_VAR_NAME=value'"
        raise click.BadParameter(msg)


def _translate_to_obj(value_list):
    apps = []
    components = []
    select_all = False
    for value in value_list:
        if value == "all":
            # all is a special keyword
            select_all = True
        elif value.startswith("app:"):
            apps.append(value.split(":")[1])
        else:
            components.append(value)
    return AppOrComponentSelector(select_all, apps, components)


def _app_or_component_selector(ctx, param, this_value):
    if any([val.startswith("-") for val in this_value]):
        raise click.BadParameter("requires a component name or keyword 'all'")

    # check if 'app:' syntax has been used and translate input values to apps/components dictionary
    this_value = _translate_to_obj(this_value)

    opposite_option = {
        "remove_resources": "no_remove_resources",
        "no_remove_resources": "remove_resources",
        "remove_dependencies": "no_remove_dependencies",
        "no_remove_dependencies": "remove_dependencies",
    }

    this_param_name: str = param.name
    other_param_name: str = opposite_option[this_param_name]

    other_value = ctx.params.get(other_param_name, AppOrComponentSelector())

    # validate that opposing options are not both set to 'all'
    if this_value.select_all and other_value.select_all:
        raise click.BadParameter(
            f"'all' cannot be specified on both this option"
            f" and its opposite '{other_param_name}'"
        )

    # validate that the same app was not used in opposing options
    for app in this_value.apps:
        if app in other_value.apps:
            raise click.BadParameter(
                f"app '{app}' cannot be specified on both this option"
                f" and its opposite '{other_param_name}'"
            )

    # validate that the same component was not used in opposing options
    for component in this_value.components:
        if component in other_value.components:
            raise click.BadParameter(
                f"component '{component}' cannot be specified on both this option"
                f" and its opposite '{other_param_name}'"
            )

    # set default value for --remove-resources to 'all' if option was unspecified
    # set default value for --no-remove-dependencies to 'all' if option was unspecified
    options_w_defaults = ("remove_resources", "no_remove_dependencies")
    if this_param_name in options_w_defaults and this_value.empty and other_value.empty:
        this_value.select_all = True

    return this_value


_app_source_options = [
    click.option(
        "--source",
        "-s",
        help=f"Configuration source to use when fetching app templates (default: {APP_SRE_SRC})",
        type=click.Choice([APP_SRE_SRC, FILE_SRC], case_sensitive=False),
        default=APP_SRE_SRC,
    ),
    click.option(
        "--local-config-path",
        "-c",
        help="File to use for local config (default: $XDG_CONFIG_HOME/bonfire/config.yaml)",
        default=None,
    ),
    click.option(
        "--local-config-method",
        help="Selects method used when combining apps in local config with remote app config",
        type=click.Choice(["merge", "override"], case_sensitive=False),
        show_default=True,
        default="merge",
    ),
    click.option(
        "--target-env",
        help=(
            f"Target environment name when using source={APP_SRE_SRC}. "
            f"Use to select which template parameters are fetched."
        ),
        type=str,
        default=conf.EPHEMERAL_ENV_NAME,
        show_default=True,
    ),
    click.option(
        "--prefer",
        "preferred_params",
        help=(
            f"When there are multiple deployment targets found in {APP_SRE_SRC}, prefer the ones "
            "that have '<parameter name>=<value>' set. Can be specified multiple times."
        ),
        multiple=True,
        default=conf.BONFIRE_DEFAULT_PREFER,
        callback=_validate_split_equals,
    ),
]

_process_options = _app_source_options + [
    click.argument(
        "app_names",
        required=True,
        nargs=-1,
    ),
    click.option(
        "--ref-env",
        help=(
            f"Reference environment name in {APP_SRE_SRC}. "
            "Use to set default 'ref'/'IMAGE_TAG' for apps."
        ),
        type=str,
        default=conf.BONFIRE_DEFAULT_REF_ENV,
    ),
    click.option(
        "--fallback-ref-env",
        help=(
            f"Reference environment name in {APP_SRE_SRC} to be used if deployment configuration"
            " not found in primary reference environment."
        ),
        type=str,
        default=conf.BONFIRE_DEFAULT_FALLBACK_REF_ENV,
    ),
    click.option(
        "--set-image-tag",
        "-i",
        help=("Override image tag for an image using format '<image uri>=<tag>'"),
        multiple=True,
        callback=_validate_split_equals,
    ),
    click.option(
        "--set-template-ref",
        help="Override template ref for a component using format '<component>=<ref>'",
        multiple=True,
        callback=_validate_set_template_ref,
    ),
    click.option(
        "--set-parameter",
        "-p",
        help=(
            "Override parameter for a component using format "
            "'<component>/<parameter name>=<value>'"
        ),
        multiple=True,
        callback=_validate_set_parameter,
    ),
    click.option(
        "--clowd-env",
        "-e",
        help=(
            "Name of ClowdEnvironment (default: if --namespace provided, will try to find match)"
        ),
        type=str,
        default=None,
    ),
    click.option(
        "--get-dependencies/--no-get-dependencies",
        help="Recursively fetch dependencies listed in ClowdApps (default: true)",
        default=True,
    ),
    click.option(
        "--optional-deps-method",
        help="Selects which method to use when fetching optionalDependencies",
        type=click.Choice(["hybrid", "all", "none"], case_sensitive=False),
        show_default=True,
        default="hybrid",
    ),
    click.option(
        "--remove-resources",
        help=(
            "Remove untrusted (defined in README) resource limits/requests on "
            "ClowdApp/ClowdJob/CJI objects for specific components or apps. Prefix the app name "
            "with 'app:', otherwise specify the component name. (default: 'all')"
        ),
        type=str,
        multiple=True,
        callback=_app_or_component_selector,
    ),
    click.option(
        "--no-remove-resources",
        help=(
            "Preserve resource limits/requests even if untrusted (defined in README) on "
            "ClowdApp/ClowdJob/CJI objects for specific components or apps. Prefix the app name "
            "with 'app:', otherwise specify the component name. (default: none)"
        ),
        type=str,
        multiple=True,
        callback=_app_or_component_selector,
    ),
    click.option(
        "--remove-dependencies",
        help=(
            "Remove dependencies/optionalDependencies on ClowdApp configs "
            "for specific components or apps. Prefix the app name with "
            "'app:', otherwise specify the component name. (default: none)"
        ),
        type=str,
        multiple=True,
        callback=_app_or_component_selector,
    ),
    click.option(
        "--no-remove-dependencies",
        help=(
            "Don't remove dependencies/optionalDependencies on ClowdApp configs "
            "for specific components or apps. Prefix the app name with "
            "'app:', otherwise specify the component name. (default: all)"
        ),
        type=str,
        multiple=True,
        callback=_app_or_component_selector,
    ),
    click.option(
        "--single-replicas/--no-single-replicas",
        help="Set replicas to '1' on all on ClowdApp configs (default: true)",
        default=True,
    ),
    click.option(
        "--component",
        "-C",
        "component_filter",
        help="Specific component(s) that should be processed (default: all)",
        type=str,
        multiple=True,
    ),
    click.option(
        "--frontends",
        "-F",
        help="Deploy frontends (default: false)",
        type=bool,
        default=False,
    ),
    _local_option,
]


_clowdenv_process_options = [
    click.option(
        "--namespace",
        "-n",
        help="Target namespace of the ClowdEnvironment",
        type=str,
        required=True,
    ),
    click.option(
        "--quay-user",
        "-u",
        help="Quay username for pullSecret provider",
        type=str,
    ),
    click.option(
        "--clowd-env",
        "-e",
        help=("Name of ClowdEnvironment (default: env-<namespace>)"),
        type=str,
        default=None,
    ),
    click.option(
        "--template-file",
        help=(
            "Path to ClowdEnvironment template file (default: use local cluster template packaged"
            " with bonfire)"
        ),
        type=str,
        default=None,
    ),
    _local_option,
]


_iqe_cji_process_options = [
    click.argument(
        "clowd_app_name",
        type=str,
        required=True,
    ),
    click.option(
        "--debug-pod",
        "debug",
        help="Set debug mode on IQE pod",
        default=False,
        is_flag=True,
    ),
    click.option(
        "--marker",
        "-m",
        help="pytest marker expression",
        type=str,
        default="",
    ),
    click.option(
        "--plugins",
        "-p",
        help="comma,separated,list of IQE plugins (default: determined by ClowdApp)",
        type=str,
        default="",
    ),
    click.option(
        "--filter",
        "-k",
        help="pytest filter expression",
        type=str,
        default="",
    ),
    click.option(
        "--env",
        "-e",
        help="dynaconf env name",
        type=str,
        default="clowder_smoke",
    ),
    click.option(
        "--image-tag",
        "-i",
        help="image tag to use for IQE pod",
        type=str,
        default="",
    ),
    click.option(
        "--selenium",
        "-s",
        help="deploy selenium container (default: false)",
        is_flag=True,
        default=False,
    ),
    click.option(
        "--cji-name",
        "-c",
        help="Name of ClowdJobInvocation (default: generate a random name)",
        type=str,
        default=None,
    ),
    click.option(
        "--template-file",
        help=(
            "Path to ClowdJobInvocation template file (default: use IQE CJI template packaged"
            " with bonfire)"
        ),
        type=str,
        default=None,
    ),
    click.option(
        "--requirements",
        help="iqe --requirements expression",
        type=str,
        default="",
    ),
    click.option(
        "--requirements-priority",
        help="iqe --requirements-priority expression",
        type=str,
        default="",
    ),
    click.option(
        "--test-importance",
        help="iqe --test-importance expression",
        type=str,
        default="",
    ),
    click.option(
        "--parallel-enabled",
        help="iqe --parallel-enabled expression",
        type=str,
        default="",
    ),
    click.option(
        "--parallel-worker-count",
        help="iqe --parallel-worker-count expression",
        type=str,
        default="",
    ),
    click.option(
        "--rp-args",
        help="iqe --rp-args expression",
        type=str,
        default="",
    ),
    click.option(
        "--ibutsu-source",
        help="iqe --ibutsu-source expression",
        type=str,
        default="",
    ),
    click.option(
        "--env-var",
        "custom_env_vars",
        help=(
            "Define a custom env var to set on the IQE pod in the format of ENV_VAR_NAME=value "
            "(can be specified multiple times)"
        ),
        multiple=True,
        callback=_validate_split_equals,
    ),
    _local_option,
]


def options(options_list):
    """Click decorator used to set a list of click options on a command."""

    def inner(func):
        for option in reversed(options_list):
            func = option(func)
        return func

    return inner


@namespace.command("list")
@options(_ns_list_options)
def _list_namespaces(available, mine, output):
    """Get list of ephemeral namespaces"""
    if not has_ns_operator():
        _error(NO_RESERVATION_SYS)

    namespaces = get_namespaces(available=available, mine=mine)

    if not namespaces:
        if output == "json":
            click.echo("{}")
        else:
            click.echo("no namespaces found")
    else:
        if output == "json":
            data = {}
            for ns in namespaces:
                data[ns.name] = {
                    "reserved": ns.reserved,
                    "status": ns.status,
                    "requester": ns.requester,
                    "expires_in": ns.expires_in,
                    "pool_type": ns.pool_type,
                }
            click.echo(json.dumps(data, indent=2))
        else:
            data = {
                "NAME": [ns.name for ns in namespaces],
                "RESERVED": [str(ns.reserved).lower() for ns in namespaces],
                "ENV STATUS": [str(ns.status).lower() for ns in namespaces],
                "APPS READY": [ns.clowdapps for ns in namespaces],
                "REQUESTER": [ns.requester for ns in namespaces],
                "POOL TYPE": [ns.pool_type for ns in namespaces],
                "EXPIRES IN": [ns.expires_in for ns in namespaces],
            }
            tabulated = tabulate(data, headers="keys")
            click.echo(tabulated)


@namespace.command("reserve")
@options(_ns_reserve_options)
@options(_timeout_option)
@click_exception_wrapper("namespace reserve")
def _cmd_namespace_reserve(name, requester, duration, pool, timeout, local, force):
    """Reserve an ephemeral namespace"""
    ns = _check_and_reserve_namespace(name, requester, duration, pool, timeout, local, force)
    click.echo(ns.name)


@namespace.command("release")
@click.argument("namespace", required=False, type=str)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Do not ask for confirmation",
)
@options([_local_option])
@click_exception_wrapper("namespace release")
def _cmd_namespace_release(namespace, force, local):
    """Remove reservation from an ephemeral namespace"""
    if not has_ns_operator():
        _error(NO_RESERVATION_SYS)

    if not namespace:
        namespace = current_namespace_or_error()

    if not force:
        ns = Namespace(name=namespace)
        if not ns.owned_by_me:
            _warn_if_not_owned_by_me()
        _warn_before_delete()

    release_reservation(namespace=namespace, local=local)


@namespace.command("extend")
@click.argument("namespace", required=False, type=str)
@click.option(
    "--duration",
    "-d",
    type=str,
    default="1h",
    help="Amount of time to extend the reservation",
    callback=_validate_reservation_duration,
)
@options([_local_option])
@click_exception_wrapper("namespace extend")
def _cmd_namespace_extend(namespace, duration, local):
    """Extend a reservation of an ephemeral namespace"""
    if not has_ns_operator():
        _error(NO_RESERVATION_SYS)

    if not namespace:
        namespace = current_namespace_or_error()

    extend_namespace(namespace, duration, local)


@namespace.command("wait-on-resources")
@click.argument("namespace", required=False, type=str)
@click.option(
    "--db-only",
    is_flag=True,
    default=False,
    help="Only wait for DB resources owned by ClowdApps to be ready",
)
@options(_timeout_option)
def _cmd_namespace_wait_on_resources(namespace, timeout, db_only):
    """Wait for rolled out resources to be ready in namespace"""
    if not namespace:
        namespace = current_namespace_or_error()
    try:
        _wait_on_namespace_resources(namespace, timeout, db_only=db_only)
    except TimedOutError as err:
        log.error("Hit timeout error: %s", err)
        _error("namespace wait timed out")


@namespace.command("describe")
@click.argument("namespace", required=False, type=str)
@click.option(
    "--output",
    "-o",
    default="cli",
    help="which output format to return the data in",
    type=click.Choice(["cli", "json"], case_sensitive=False),
)
def _describe_namespace(namespace, output):
    """Get current namespace info"""
    if not namespace:
        namespace = current_namespace_or_error()

    click.echo(describe_namespace(namespace, output))


def _get_apps_config(
    source,
    target_env,
    ref_env,
    fallback_ref_env,
    local_config_path,
    local_config_method,
    preferred_params,
):
    config = conf.load_config(local_config_path)

    if source == APP_SRE_SRC:
        log.info("fetching target env apps config using source: %s", source)
        if not target_env:
            _error("target env must be supplied for source '{APP_SRE_SRC}'")
        apps_config = get_apps_for_env(target_env, preferred_params)

        if not ref_env and target_env == conf.EPHEMERAL_ENV_NAME:
            # set git target to 'master' because ephemeral targets have no git ref defined
            log.info(
                "target env is '%s' and no ref env given, using 'master' git ref for all apps",
                conf.EPHEMERAL_ENV_NAME,
            )
            for _, app_cfg in apps_config.items():
                for component in app_cfg.get("components", []):
                    component["ref"] = "master"

    elif source == FILE_SRC:
        log.info("fetching apps config using source: %s", source)
        apps_config = get_appsfile_apps(config)

    # merge remote apps config with local app config
    local_apps = get_local_apps(config)
    apps_config = merge_app_configs(apps_config, local_apps, local_config_method)

    # validate the components look ok after merging
    for app_name, app_config in apps_config.items():
        for component in app_config["components"]:
            # validate the config for a component
            if not component.get("name"):
                raise FatalError(f"{SYNTAX_ERR}, component is missing 'name'")
            try:
                RepoFile.from_config(component)
            except FatalError as err:
                # re-raise with a bit more context
                raise FatalError(f"{str(err)}, hit on app {app_name}")

    # handle git ref/image substitutions if reference environment was provided
    if ref_env:
        apps_config = sub_refs(apps_config, ref_env, fallback_ref_env, preferred_params)

    return apps_config


def _log_and_return(env_name):
    log.info("templates will be processed with parameter ENV_NAME='%s'", env_name)
    return env_name


def _get_env_name(ns=None, env_name=None):
    if env_name:
        return _log_and_return(env_name)

    if not ns:
        log.warning("neither '--clowd-env' nor '--namespace' provided")
        return _log_and_return(None)

    log.info("searching for ClowdEnvironment tied to ns '%s'...", ns)
    match = find_clowd_env_for_ns(ns)
    if not match:
        log.warning(
            "could not find a ClowdEnvironment with target ns '%s'.  "
            "Specify one with '--clowd-env' if needed.",
            ns,
        )
        return _log_and_return(None)

    return _log_and_return(match["metadata"]["name"])


def _process(
    app_names,
    source,
    get_dependencies,
    optional_deps_method,
    local_config_method,
    set_image_tag,
    ref_env,
    fallback_ref_env,
    target_env,
    set_template_ref,
    set_parameter,
    clowd_env,
    local_config_path,
    remove_resources,
    no_remove_resources,
    remove_dependencies,
    no_remove_dependencies,
    single_replicas,
    component_filter,
    local,
    frontends,
    preferred_params,
):
    apps_config = _get_apps_config(
        source,
        target_env,
        ref_env,
        fallback_ref_env,
        local_config_path,
        local_config_method,
        preferred_params,
    )

    processor = TemplateProcessor(
        apps_config,
        app_names,
        get_dependencies,
        optional_deps_method,
        set_image_tag,
        set_template_ref,
        set_parameter,
        clowd_env,
        remove_resources,
        no_remove_resources,
        remove_dependencies,
        no_remove_dependencies,
        single_replicas,
        component_filter,
        local,
        frontends,
    )
    return processor.process()


@pool.command("list")
def _cmd_pool_types():
    """List all pool types"""
    click.echo("\n".join(get_namespace_pools()))


def _get_return_args(*args, **kwargs):
    """Dummy function used for unit testing process options"""
    pass


@test.command("process", hidden=True)
@options(_process_options)
def _cmd_test_process(*args, **kwargs):
    """Dummy command used for unit testing process options"""
    _get_return_args(*args, **kwargs)


@main.command("process")
@options(_process_options)
@click.option(
    "--namespace",
    "-n",
    help="Namespace you intend to deploy to (default: none)",
    type=str,
)
def _cmd_process(
    app_names,
    source,
    get_dependencies,
    optional_deps_method,
    local_config_method,
    set_image_tag,
    ref_env,
    fallback_ref_env,
    target_env,
    set_template_ref,
    set_parameter,
    clowd_env,
    namespace,
    local_config_path,
    remove_resources,
    no_remove_resources,
    remove_dependencies,
    no_remove_dependencies,
    single_replicas,
    component_filter,
    local,
    frontends,
    preferred_params,
):
    """Fetch and process application templates"""
    clowd_env = _get_env_name(namespace, clowd_env)

    processed_templates = _process(
        app_names,
        source,
        get_dependencies,
        optional_deps_method,
        local_config_method,
        set_image_tag,
        ref_env,
        fallback_ref_env,
        target_env,
        set_template_ref,
        set_parameter,
        clowd_env,
        local_config_path,
        remove_resources,
        no_remove_resources,
        remove_dependencies,
        no_remove_dependencies,
        single_replicas,
        component_filter,
        local,
        frontends,
        preferred_params,
    )
    print(json.dumps(processed_templates, indent=2))


def _get_namespace(
    requested_ns_name, name, requester, duration, pool, timeout, local, force, using_current=False
):
    if not has_ns_operator():
        if requested_ns_name:
            ns = Namespace(name=requested_ns_name)
            return ns.name, False
        else:
            _error(f"{NO_RESERVATION_SYS}. Use '-n' to provide a specific target namespace")

    ns = None
    if requested_ns_name:
        ns = _check_and_use_namespace(requested_ns_name, using_current, requester)

    reserved_new_ns = False
    if not ns:
        if using_current:
            log.info(
                "current namespace could not be used (not reserved,"
                " expired, or not owned), reserving a new one",
            )
        ns = _check_and_reserve_namespace(name, requester, duration, pool, timeout, local, force)
        reserved_new_ns = True

    return ns.name, reserved_new_ns


def _check_and_use_namespace(requested_ns_name, using_current, requester):
    if using_current:
        log.info("attempting to use current namespace from oc/kubectl context...")

    if not has_ns_operator():
        _error(f"{NO_RESERVATION_SYS}")

    log.debug("checking if namespace '%s' has been reserved via ns operator...", requested_ns_name)
    operator_reservation = get_reservation(namespace=requested_ns_name)
    ns = None
    if operator_reservation:
        log.debug("found existing ns operator reservation")

        if operator_reservation.get("status", {}).get("state") == "expired":
            msg = f"reservation has expired for namespace '{requested_ns_name}'"
            if using_current:
                log.info(msg)
                return None
            else:
                _error(msg)

        ns = Namespace(name=requested_ns_name)

        if ns.owned_by_me or ns.requester == requester:
            log.info("namespace '%s' is owned by this user", ns.name)
        else:
            log.info("namespace '%s' is reserved by someone else", ns.name)
            if using_current:
                return None
            _warn_if_not_owned_by_me()

        if not ns.ready:
            _warn_if_not_ready()

    elif not using_current:
        _error(f"Namespace '{requested_ns_name}' has not been reserved yet")

    return ns


def _check_and_reserve_namespace(name, requester, duration, pool, timeout, local, force):
    if not has_ns_operator():
        _error(f"{NO_RESERVATION_SYS}")

    if pool not in get_namespace_pools():
        _error(f"namespace pool '{pool}' does not exist on this cluster")

    log.debug("checking if requester already has another namespace reserved...")
    requester = requester if requester else _get_requester()
    if not force and check_for_existing_reservation(requester):
        _warn_of_existing(requester)

    log.info("checking for available namespaces to reserve...")

    pool_size_limit = get_pool_size_limit(pool)
    log.info("pool size limit is defined as %d in '%s' pool", pool_size_limit, pool)
    if pool_size_limit > 0 and get_reserved_namespace_quantity(pool) >= pool_size_limit:
        _error(
            f"Maximum number of namespaces for pool `{pool}` (limit: {pool_size_limit})"
            " have been reserved"
        )

    return reserve_namespace(name, requester, duration, pool, timeout, local)


@main.command("deploy")
@options(_process_options)
@click.option(
    "--namespace",
    "-n",
    help=(
        "Namespace (defaults to namespace from current context; "
        "if not set, not reserved, or not owned, then bonfire will reserve a new one)"
    ),
    default=None,
)
@click.option(
    "--reserve",
    help=(
        "Do not use current context's namespace and force reserve a new one. (keeps"
        " the reservation on failure)"
    ),
    is_flag=True,
)
@click.option(
    "--import-secrets",
    is_flag=True,
    help="Import secrets from local directory at deploy time",
    default=False,
)
@click.option(
    "--import-configmaps",
    is_flag=True,
    help="Import configmaps from local directory at deploy time",
    default=False,
)
@click.option(
    "--secrets-dir",
    type=str,
    help="Directory to use for secrets import (default: $XDG_CONFIG_HOME/bonfire/secrets/)",
    default=conf.DEFAULT_SECRETS_DIR,
)
@click.option(
    "--configmaps-dir",
    type=str,
    help=("Directory to use for configmaps import (default: $XDG_CONFIG_HOME/bonfire/configmaps/)"),
    default=conf.DEFAULT_CONFIGMAPS_DIR,
)
@click.option(
    "--no-release-on-fail",
    is_flag=True,
    help="Do not release namespace reservation if deployment fails",
)
@options(_ns_reserve_options)
@options(_timeout_option)
def _cmd_config_deploy(
    app_names,
    source,
    get_dependencies,
    optional_deps_method,
    local_config_method,
    set_image_tag,
    ref_env,
    fallback_ref_env,
    target_env,
    set_template_ref,
    set_parameter,
    clowd_env,
    local_config_path,
    remove_resources,
    no_remove_resources,
    remove_dependencies,
    no_remove_dependencies,
    single_replicas,
    namespace,
    reserve,
    name,
    requester,
    duration,
    timeout,
    no_release_on_fail,
    component_filter,
    import_secrets,
    import_configmaps,
    secrets_dir,
    configmaps_dir,
    local,
    frontends,
    pool,
    force,
    preferred_params,
):
    """Process app templates and deploy them to a cluster"""
    if not has_clowder():
        _error("cluster does not have clowder operator installed")

    using_current = False
    if reserve:
        namespace = None
    elif not namespace and not conf.BONFIRE_BOT:
        using_current = True
        namespace = get_current_namespace()

    ns, reserved_new_ns = _get_namespace(
        namespace,
        name,
        requester,
        duration,
        pool,
        timeout,
        local,
        force,
        using_current=using_current,
    )

    if import_secrets:
        import_secrets_from_dir(secrets_dir)

    if import_configmaps:
        import_configmaps_from_dir(configmaps_dir)

    clowd_env = _get_env_name(ns, clowd_env)

    def _err_handler(err):
        try:
            if not no_release_on_fail and reserved_new_ns and not reserve:
                # if we auto-reserved this ns, auto-release it on failure unless
                # --no-release-on-fail was requested
                log.info("releasing namespace '%s'", ns)
                release_reservation(namespace=ns)
        finally:
            msg = "deploy failed"
            if str(err):
                msg += f": {str(err)}"
            _error(msg)

    try:
        log.info("processing app templates...")
        apps_config = _process(
            app_names,
            source,
            get_dependencies,
            optional_deps_method,
            local_config_method,
            set_image_tag,
            ref_env,
            fallback_ref_env,
            target_env,
            set_template_ref,
            set_parameter,
            clowd_env,
            local_config_path,
            remove_resources,
            no_remove_resources,
            remove_dependencies,
            no_remove_dependencies,
            single_replicas,
            component_filter,
            local,
            frontends,
            preferred_params,
        )
        log.debug("app configs:\n%s", json.dumps(apps_config, indent=2))
        if not apps_config["items"]:
            log.warning("no configurations found to apply!")
        else:
            log.info("applying app configs...")
            apply_config(ns, apps_config)
            log.info("waiting on resources for max of %dsec...", timeout)
            _wait_on_namespace_resources(ns, timeout)
    except KeyboardInterrupt as err:
        log.error("aborted by keyboard interrupt!")
        _err_handler(err)
    except TimedOutError as err:
        log.error("hit timeout error: %s", err)
        _err_handler(err)
    except FatalError as err:
        log.error("hit fatal error: %s", err)
        _err_handler(err)
    except StatusError as err:
        log.error("hit status error: %s", err)
        _err_handler(err)
    except Exception as err:
        log.exception("hit unexpected error!")
        _err_handler(err)
    else:
        log.info("successfully deployed to namespace %s", ns)
        es_telemetry.send_telemetry("successful deployment")
        url = get_console_url()
        if url:
            ns_url = f"{url}/k8s/cluster/projects/{ns}"
            log.info("namespace url: %s", ns_url)
            log.info(
                "resource usage dashboard for namespace '%s': %s",
                ns,
                conf.RESOURCE_DASHBOARD_URL.format(namespace=ns),
            )
        click.echo(ns)


def _process_clowdenv(namespace, quay_user, clowd_env, template_file, local):
    if not clowd_env:
        clowd_env = f"env-{namespace}"
    return process_clowd_env(namespace, quay_user, clowd_env, template_file, local)


@main.command("process-env")
@options(_clowdenv_process_options)
def _cmd_process_clowdenv(namespace, quay_user, clowd_env, template_file, local):
    """Process ClowdEnv template and print output"""
    clowd_env_config = _process_clowdenv(namespace, quay_user, clowd_env, template_file, local)
    print(json.dumps(clowd_env_config, indent=2))


@main.command("deploy-env")
@options(_clowdenv_process_options)
@click.option(
    "--import-secrets",
    is_flag=True,
    help="Import secrets from local directory at deploy time",
    default=False,
)
@click.option(
    "--import-configmaps",
    is_flag=True,
    help="Import configmaps from local directory at deploy time",
    default=False,
)
@click.option(
    "--secrets-dir",
    type=str,
    help=("Import secrets from this directory (default: " "$XDG_CONFIG_HOME/bonfire/secrets/)"),
    default=conf.DEFAULT_SECRETS_DIR,
)
@click.option(
    "--configmaps-dir",
    type=str,
    help=(
        "Import configmaps from this directory \
           (default: "
        "$XDG_CONFIG_HOME/bonfire/configmaps/)"
    ),
    default=conf.DEFAULT_CONFIGMAPS_DIR,
)
@options(_ns_reserve_options)
@options(_timeout_option)
@click_exception_wrapper("deploy-env")
def _cmd_deploy_clowdenv(
    namespace,
    quay_user,
    clowd_env,
    template_file,
    timeout,
    import_secrets,
    import_configmaps,
    secrets_dir,
    configmaps_dir,
    name,
    requester,
    duration,
    local,
    pool,
    force,
):
    """Process ClowdEnv template and deploy to a cluster"""
    if not has_clowder():
        _error("cluster does not have clowder operator installed")

    namespace, _ = _get_namespace(namespace, name, requester, duration, pool, timeout, local, force)

    if import_secrets:
        import_secrets_from_dir(secrets_dir)

    if import_configmaps:
        import_configmaps_from_dir(configmaps_dir)

    clowd_env_config = _process_clowdenv(namespace, quay_user, clowd_env, template_file, local)

    log.debug("ClowdEnvironment config:\n%s", clowd_env_config)

    apply_config(None, clowd_env_config)

    if not namespace:
        # wait for Clowder to tell us what target namespace it created
        namespace = wait_for_clowd_env_target_ns(clowd_env)

    log.info("waiting on resources for max of %dsec...", timeout)
    _wait_on_namespace_resources(namespace, timeout)

    clowd_env_name = find_clowd_env_for_ns(namespace)["metadata"]["name"]

    log.info("ClowdEnvironment '%s' using ns '%s' is ready", clowd_env_name, namespace)
    click.echo(namespace)


@main.command("process-iqe-cji")
@options(_iqe_cji_process_options)
def _cmd_process_iqe_cji(
    clowd_app_name,
    debug,
    marker,
    filter,
    env,
    image_tag,
    cji_name,
    template_file,
    requirements,
    requirements_priority,
    test_importance,
    plugins,
    local,
    selenium,
    parallel_enabled,
    parallel_worker_count,
    rp_args,
    ibutsu_source,
    custom_env_vars,
):
    """Process IQE ClowdJobInvocation template and print output"""
    cji_config = process_iqe_cji(
        clowd_app_name,
        debug,
        marker,
        filter,
        env,
        image_tag,
        cji_name,
        template_file,
        requirements,
        requirements_priority,
        test_importance,
        plugins,
        local,
        selenium,
        parallel_enabled,
        parallel_worker_count,
        rp_args,
        ibutsu_source,
        custom_env_vars,
    )
    print(json.dumps(cji_config, indent=2))


@main.command("deploy-iqe-cji")
@click.option("--namespace", "-n", help="Namespace to deploy to", type=str, required=True)
@options(_iqe_cji_process_options)
@options(_ns_reserve_options)
@options(_timeout_option)
@click_exception_wrapper("deploy-iqe-cji")
def _cmd_deploy_iqe_cji(
    namespace,
    clowd_app_name,
    debug,
    marker,
    filter,
    env,
    image_tag,
    cji_name,
    template_file,
    timeout,
    requirements,
    requirements_priority,
    test_importance,
    plugins,
    name,
    requester,
    duration,
    local,
    selenium,
    parallel_enabled,
    parallel_worker_count,
    rp_args,
    ibutsu_source,
    custom_env_vars,
    pool,
    force,
):
    """Process IQE CJI template, apply it, and wait for it to start running."""
    if not has_clowder():
        _error("cluster does not have clowder operator installed")

    namespace, _ = _get_namespace(namespace, name, requester, duration, pool, timeout, local, force)

    cji_config = process_iqe_cji(
        clowd_app_name,
        debug,
        marker,
        filter,
        env,
        image_tag,
        cji_name,
        template_file,
        requirements,
        requirements_priority,
        test_importance,
        plugins,
        local,
        selenium,
        parallel_enabled,
        parallel_worker_count,
        rp_args,
        ibutsu_source,
        custom_env_vars,
    )

    log.debug("processed CJI config:\n%s", cji_config)

    try:
        cji_name = cji_config["items"][0]["metadata"]["name"]
    except (KeyError, IndexError):
        raise Exception("error parsing name of CJI from processed template, check CJI template")

    apply_config(namespace, cji_config)

    log.info("waiting on CJI '%s' for max of %dsec...", cji_name, timeout)
    pod_name = wait_on_cji(namespace, cji_name, timeout)
    log.info("pod '%s' related to CJI '%s' in ns '%s' is running", pod_name, cji_name, namespace)
    click.echo(pod_name)


@main.command("version")
def _cmd_version():
    """Print bonfire version"""
    click.echo("bonfire version " + get_version())


@config.command("write-default")
@click.argument("path", required=False, type=str)
def _cmd_write_default_config(path):
    """Write default configuration file to PATH (default: $XDG_CONFIG_HOME/bonfire/config.yaml)"""
    conf.write_default_config(path)


@config.command("edit")
@click.argument("path", required=False, type=str)
def _cmd_edit_default_config(path):
    """Edit configuration with $EDITOR (default path: $XDG_CONFIG_HOME/bonfire/config.yaml)"""
    conf.edit_default_config(path)


@options(_app_source_options)
@click.option(
    "--components/--no-components",
    "list_components",
    default=False,
    help="List components contained within each app group",
)
@apps.command("list")
def _cmd_apps_list(
    source,
    local_config_path,
    local_config_method,
    target_env,
    list_components,
    preferred_params,
):
    """List names of all apps that are marked for deployment in given 'target_env'"""
    apps = _get_apps_config(
        source, target_env, None, None, local_config_path, local_config_method, preferred_params
    )

    print("")
    sorted_keys = sorted(apps.keys())
    for app_name in sorted_keys:
        app_config = apps[app_name]
        print(app_name)
        if list_components:
            component_names = sorted([c["name"] for c in app_config["components"]])
            for component_name in component_names:
                print(f" `-- {component_name}")


@options(_app_source_options)
@click.argument(
    "component",
    type=str,
)
@apps.command("what-depends-on")
def _cmd_apps_what_depends_on(
    source,
    local_config_path,
    local_config_method,
    target_env,
    component,
    preferred_params,
):
    """Show any apps that depend on COMPONENT for deployments in given 'target_env'"""
    apps = _get_apps_config(
        source, target_env, None, None, local_config_path, local_config_method, preferred_params
    )
    found = find_what_depends_on(apps, component)
    print("\n".join(found) or f"no apps depending on {component} found")


def main_with_handler():
    try:
        main()
    except FatalError as err:
        _error(str(err))


if __name__ == "__main__":
    main_with_handler()



import logging

import yaml

from bonfire.utils import FatalError, RepoFile, get_dupes, SYNTAX_ERR

log = logging.getLogger(__name__)


def _fetch_apps_file(config):
    rf = RepoFile.from_config(config["appsFile"])
    commit, content = rf.fetch()
    log.info(
        "loading commit '%s' of %s repo %s/%s at path '%s' for apps config",
        commit,
        rf.host,
        rf.org,
        rf.repo,
        rf.path,
    )
    fetched_apps = yaml.safe_load(content)

    if "apps" not in fetched_apps:
        raise FatalError(f"{SYNTAX_ERR}, fetched apps file has no 'apps' key")

    app_names = [a["name"] for a in fetched_apps["apps"]]
    dupes = get_dupes(app_names)
    if dupes:
        raise FatalError(f"{SYNTAX_ERR}, duplicate app names found in fetched apps file: {dupes}")

    return {a["name"]: a for a in fetched_apps["apps"]}


def _parse_apps_in_cfg(config):
    app_names = []

    for app in config["apps"]:
        if not isinstance(app, dict):
            raise FatalError(f"{SYNTAX_ERR} app type should be a dict type")
        if "components" not in app:
            raise FatalError(f"{SYNTAX_ERR} 'components' missing from an app")
        for component in app["components"]:
            if not isinstance(component, dict):
                raise FatalError(f"{SYNTAX_ERR} component should be a dict type")
        app_names.append(app["name"])

    dupes = get_dupes(app_names)
    if dupes:
        raise FatalError(f"{SYNTAX_ERR} duplicate app names found in config: {dupes}")

    return {a["name"]: a for a in config["apps"]}


def get_local_apps(config):
    """
    Get apps defined locally under 'apps' section of config
    """
    config_apps = {}
    if not isinstance(config, dict):
        raise FatalError(f"{SYNTAX_ERR}, expected local config to be a dictionary")
    if "apps" in config:
        config_apps = _parse_apps_in_cfg(config)
        log.info("local configuration found for apps: %s", list(config_apps.keys()))

    return config_apps


def get_appsfile_apps(config):
    """
    Fetch apps from repo based on appsFile provided in config
    """
    if not isinstance(config, dict):
        raise FatalError(f"{SYNTAX_ERR}, expected local config to be a dictionary")

    if "appsFile" not in config:
        raise FatalError(f"{SYNTAX_ERR}, config has no 'appsFile' defined")

    log.info("local config has a remote 'appsFile' defined, fetching it...")
    fetched_apps = _fetch_apps_file(config)
    return fetched_apps



import copy
import json
import logging
import re
from urllib.parse import urlparse

from gql import Client as GQLClient
from gql import gql
from gql.transport.requests import RequestsHTTPTransport
from requests.auth import HTTPBasicAuth

import bonfire.config as conf
from bonfire.utils import check_url_connection

log = logging.getLogger(__name__)


CONSOLEDOT_PARENT_APPS = ("insights", "image-builder")


ENVS_QUERY = gql(
    """
    {
      envs: environments_v1 {
        name
        parameters
        namespaces {
          name
          path
        }
      }
    }
    """
)

APPS_QUERY = gql(
    """
    {
      apps: apps_v1 {
        name
        parentApp {
          name
        }
        saasFiles {
          path
          name
          parameters
          resourceTemplates {
            name
            path
            url
            parameters
            targets {
              namespace {
                name
                path
                cluster {
                  name
                }
              }
              ref
              parameters
            }
          }
        }
      }
    }
    """
)


class Client:
    def __init__(self):
        log.debug("using url: %s", conf.QONTRACT_BASE_URL)

        transport_kwargs = {"url": conf.QONTRACT_BASE_URL}

        if conf.QONTRACT_TOKEN:
            log.debug("using token authentication")
            transport_kwargs["headers"] = {"Authorization": conf.QONTRACT_TOKEN}
        elif conf.QONTRACT_USERNAME and conf.QONTRACT_PASSWORD:
            log.debug("using basic authentication")
            transport_kwargs["auth"] = HTTPBasicAuth(conf.QONTRACT_USERNAME, conf.QONTRACT_PASSWORD)

        check_url_connection(transport_kwargs["url"])

        transport = RequestsHTTPTransport(**transport_kwargs)
        self.client = GQLClient(transport=transport, fetch_schema_from_transport=True)

        # info level is way too noisy for the gql client
        logging.getLogger("gql").setLevel(logging.ERROR)

    def get_env(self, env):
        """Get insights env configuration."""
        for env_data in self.client.execute(ENVS_QUERY)["envs"]:
            if env_data["name"] == env:
                env_data["namespaces"] = {
                    ns["path"]: ns["name"] for ns in env_data.get("namespaces", [])
                }
                break
        else:
            raise ValueError(f"cannot find env '{env}'")

        return env_data

    def get_apps(self):
        return self.client.execute(APPS_QUERY)["apps"]


_client = None


def get_client():
    global _client
    if not _client:
        _client = Client()

    return _client


def _to_dict(nullable_json_str):
    return json.loads(nullable_json_str or "{}")


def _find_matching_component(apps, app_name, component_name):
    for component in apps.get(app_name, {}).get("components", []):
        if component["name"] == component_name:
            return component


def _check_replace_other(other_params, this_params, preferred_params):
    """
    Compare parameters of "this" component with parameters of the "other" component.
    Assign points to the component if it is using a preferred parameter value. The component
    with the higher number of points will be selected.
    """
    this_weight = 0
    other_weight = 0

    # prefer deployment targets that have CLOWDER_ENABLED=true
    # TODO: a relic of the past and at this point probably no longer needed, remove in a follow-up
    preferred_params["CLOWDER_ENABLED"] = preferred_params.get("CLOWDER_ENABLED", "true")

    # check if target is configured with preferred parameters
    for param_name, param_value in preferred_params.items():
        if str(this_params.get(param_name)).lower() == str(param_value).lower():
            log.debug("    `-- 'this' weight +1 for %s=%s", param_name, param_value)
            this_weight += 1
        if str(other_params.get(param_name)).lower() == str(param_value).lower():
            log.debug("    `-- 'other' weight +1 for %s=%s", param_name, param_value)
            other_weight += 1

    # prefer deployment targets that have REPLICAS >= 1
    for param_name in ("REPLICAS", "MIN_REPLICAS"):
        this_replicas = int(this_params.get(param_name, 0))
        other_replicas = int(other_params.get(param_name, 0))
        if this_replicas >= 1:
            log.debug("    `-- 'this' weight +1 for %s>=1", param_name)
            this_weight += 1
        if other_replicas >= 1:
            log.debug("    `-- 'other' weight +1 for %s>=1", param_name)
            other_weight += 1

    log.debug("    `-- final: 'this' weight: %d, 'other' weight: %d", this_weight, other_weight)

    if this_weight > other_weight:
        return True

    return False


def _add_component_if_priority_higher(
    apps,
    app_name,
    component_name,
    component,
    defined_multiple,
    preferred_params,
):
    existing_match = _find_matching_component(apps, app_name, component_name)
    if not existing_match:
        apps[app_name]["name"] = app_name
        apps[app_name]["components"].append(component)
    else:
        # this app/component is defined multiple times in the environment
        # look at the parameters set on it to decide which definition to prioritize
        defined_multiple.add((app_name, component_name))
        other_params = existing_match["parameters"]
        this_params = component["parameters"]
        log.debug("  `-- this is a duplicate target, checking if this replaces existing other")

        replace = _check_replace_other(other_params, this_params, preferred_params)

        if replace:
            log.debug("  `-- this is weighted higher, replaces other")
            apps[app_name]["components"].remove(existing_match)
            apps[app_name]["components"].append(component)
        else:
            log.debug("  `-- this is weighted equal/lower, not replacing")


def _process_env_parameters(parameters):
    """Process variable reference in place, e.g. KAFKA_URL='${KAFKA_HOST}:9092'"""
    for key, val in parameters.items():
        if isinstance(val, str):
            found = re.findall(r"\$\{([^$]+)\}", val)
            for var in found:
                if var in parameters:
                    parameters[key] = parameters[key].replace("${" + var + "}", parameters[var])
                    log.debug(
                        "parameter %s found and replaced with %s",
                        var,
                        parameters[var],
                    )


def _add_component(
    apps,
    env,
    app_name,
    saas_file,
    resource_template,
    target,
    defined_multiple,
    preferred_params,
):
    component_name = resource_template["name"]
    saas_file_path = saas_file["path"]

    if app_name not in apps:
        apps[app_name] = {"name": app_name, "components": []}

    url = resource_template["url"]
    if "github" not in url and "gitlab" not in url:
        raise ValueError(
            f"unknown host for resourceTemplate url '{url}' found in saas file '{saas_file_path}'"
        )
    host = "github" if "github" in url else "gitlab"

    try:
        parsed_url = urlparse(url)
        org, repo = parsed_url.path.rstrip("/").split("/")[-2:]
    except (ValueError, IndexError) as err:
        raise ValueError(f"invalid repo url '{url}': {err}")

    # merge the various layers of parameters
    p = copy.deepcopy(_to_dict(env["parameters"]))
    p.update(_to_dict(saas_file["parameters"]))
    p.update(_to_dict(resource_template["parameters"]))
    p.update(_to_dict(target["parameters"]))
    _process_env_parameters(p)

    component = {
        "name": component_name,
        "path": resource_template["path"],
        "host": host,
        "repo": f"{org}/{repo}",
        "ref": target["ref"],
        "parameters": p,
    }

    _add_component_if_priority_higher(
        apps,
        app_name,
        component_name,
        component,
        defined_multiple,
        preferred_params,
    )


def get_apps_for_env(env_name, preferred_params):
    if not env_name:
        return {}

    log.info("fetching app deployment configs for env '%s'", env_name)

    client = get_client()
    all_apps = client.get_apps()
    env = client.get_env(env_name)

    # work-around to only show apps with an ephemeral deploy target
    if env_name == conf.EPHEMERAL_ENV_NAME:
        env["namespaces"] = [conf.BASE_NAMESPACE_PATH]

    apps = {}
    ignored_apps = set()
    defined_multiple = set()

    for app in all_apps:
        if app["parentApp"] and app["parentApp"].get("name") not in CONSOLEDOT_PARENT_APPS:
            ignored_apps.add(app["name"])
            continue
        saas_files = app.get("saasFiles", [])
        for saas_file in saas_files:
            for rt_idx, resource_template in enumerate(saas_file.get("resourceTemplates", [])):
                for target_idx, target in enumerate(resource_template.get("targets", [])):
                    ns = target.get("namespace") or {}
                    ns_name = ns.get("name")
                    ns_path = ns.get("path")
                    if ns_path not in env.get("namespaces", []):
                        # this deploy target ns is not in the environment
                        continue
                    # this target ns belongs to the environment
                    log.debug(
                        "app '%s' component '%s' found in saas file '%s'",
                        app["name"],
                        resource_template["name"],
                        saas_file["path"],
                    )
                    log.debug(
                        "  position: .resourceTemplates[%d].targets[%d] (env '%s', ns '%s')",
                        rt_idx,
                        target_idx,
                        env_name,
                        ns_name,
                    )
                    _add_component(
                        apps,
                        env,
                        app["name"],
                        saas_file,
                        resource_template,
                        target,
                        defined_multiple,
                        preferred_params,
                    )

    if ignored_apps:
        log.debug(
            "ignored apps in env '%s' that do not have parentApp of %s: %s",
            CONSOLEDOT_PARENT_APPS,
            env_name,
            ", ".join(ignored_apps),
        )

    if defined_multiple:
        log.debug(
            "the following components in env '%s' are defined multiple times: %s",
            env_name,
            ", ".join([f"{app}/{component}" for app, component in defined_multiple]),
        )

    return apps


def _find_ref_target_and_update_component(
    final_apps,
    ref_env_apps,
    fallback_ref_env_apps,
    ref_env,
    fallback_ref_env,
    app_name,
    component_idx,
    component_name,
):
    log_prefix = f"app: '{app_name}' component: '{component_name}' --"

    ref_component = _find_matching_component(ref_env_apps, app_name, component_name)

    if not ref_component and fallback_ref_env:
        log.debug(
            "%s no deploy cfg found in ref env '%s', trying fallback ref '%s'",
            log_prefix,
            ref_env,
            fallback_ref_env,
        )
        ref_component = _find_matching_component(fallback_ref_env_apps, app_name, component_name)

    if not ref_component:
        log.debug(
            "%s no deploy cfg found in ref env '%s' nor fallback '%s', using git ref 'master'",
            log_prefix,
            ref_env,
            fallback_ref_env or "(none)",
        )
        final_apps[app_name]["components"][component_idx]["ref"] = "master"

    else:
        # use git ref from reference deployment config
        final_component = final_apps[app_name]["components"][component_idx]
        final_component["ref"] = ref_component["ref"]

        # fetch any param starting with 'IMAGE_TAG' from the reference deployment config
        # and update the component's parameters with the new parameter values.
        image_tags = {}
        parameters = ref_component.get("parameters", {})
        for param, val in parameters.items():
            if param.startswith("IMAGE_TAG"):
                image_tags[param] = val
        if image_tags:
            if "parameters" not in final_component:
                final_component["parameters"] = {}
            final_component["parameters"].update(image_tags)

        log.debug(
            "%s using git ref/image tag from env '%s': %s%s",
            log_prefix,
            ref_env,
            final_component["ref"],
            f", {image_tags}" if image_tags else "",
        )


def sub_refs(apps, ref_env, fallback_ref_env=None, preferred_params=None):
    if not preferred_params:
        preferred_params = {}

    if fallback_ref_env == ref_env:
        # it would be a waste of time to try to fall back to the same env
        log.debug("ref env and fallback ref env are the same, setting fallback to 'None'")
        fallback_ref_env = None

    final_apps = copy.deepcopy(apps)
    log.info(
        "setting git refs/image tags to match deploy config found in env: %s, fallback env: %s",
        ref_env,
        fallback_ref_env or "(none)",
    )
    ref_env_apps = get_apps_for_env(ref_env, preferred_params)
    fallback_ref_env_apps = get_apps_for_env(fallback_ref_env, preferred_params)

    for app_name, app in apps.items():
        for component_idx, component in enumerate(app["components"]):
            _find_ref_target_and_update_component(
                final_apps,
                ref_env_apps,
                fallback_ref_env_apps,
                ref_env,
                fallback_ref_env,
                app_name,
                component_idx,
                component["name"],
            )

    return final_apps



import functools
import logging
import time

from ocviapy import (
    Resource,
    ResourceWaiter,
    ResourceWatcher,
    available_checkable_resources,
    get_api_resources,
    get_json,
    oc,
    wait_for_ready_threaded,
    on_k8s,
    get_all_namespaces,
)
from sh import ErrorReturnCode
from wait_for import TimedOutError, wait_for

log = logging.getLogger(__name__)


@functools.lru_cache(maxsize=None, typed=False)
def has_ns_operator():
    for res in get_api_resources():
        name = res["name"]
        apigroup = res["apigroup"].split("/")[0]
        if name == "namespacereservation" and apigroup == "cloud.redhat.com":
            return True
    return False


@functools.lru_cache(maxsize=None, typed=False)
def get_console_url():
    if on_k8s():
        return None
    else:
        try:
            cfg_map = get_json("configmap", "console-public", namespace="openshift-config-managed")
            url = cfg_map["data"]["consoleURL"]
        except Exception as err:
            log.debug("unable to obtain console url: %s: %s", err.__class__.__name__, err)
            return None
        return url


@functools.lru_cache(maxsize=None, typed=False)
def get_namespace_pools():
    namespace_pools = get_json("namespacepool")
    return [pool["metadata"]["name"] for pool in namespace_pools.get("items", [])]


def has_clowder():
    for res in get_api_resources():
        name = res["name"]
        apigroup = res["apigroup"].split("/")[0]
        if name == "clowdapp" and apigroup == "cloud.redhat.com":
            return True
    return False


# we will assume that 'oc whoami' will not change during execution
@functools.lru_cache(maxsize=None, typed=False)
def whoami():
    name = oc("whoami", _silent=True).strip()
    # a valid label must be an empty string or consist of alphanumeric characters,
    # '-', '_' or '.', and must start and end with an alphanumeric character, so let's just sanitize
    # the name at this point
    return name.replace("@", "_at_").replace(":", "_")


def _resources_for_ns_wait():
    """Only check "higher level" resource types when waiting on a namespace"""
    resources = available_checkable_resources(namespaced=True)
    try:
        resources.remove("pod")
    except ValueError as err:
        if "not in list" in str(err):
            pass
        else:
            raise
    return resources


def _all_resources_ready(namespace, timeout, watcher):
    already_waited_on = set()

    # wait on ClowdEnvironments if any ClowdApps reference one
    start = time.time()

    clowd_envs = set()
    clowdapps = get_json("clowdapp", namespace=namespace)

    for clowdapp in clowdapps["items"]:
        env_name = clowdapp["spec"]["envName"]
        if env_name not in clowd_envs:
            log.info(
                "will wait on ClowdEnvironment '%s' found on ClowdApp's .spec.envName",
                env_name,
            )
            clowd_envs.add(env_name)

    env_waiters = []
    for clowd_env in clowd_envs:
        waiter = ResourceWaiter(
            namespace,
            "clowdenvironment",
            clowd_env,
            watch_owned=True,
            watcher=watcher,
        )
        env_waiters.append(waiter)
    if not wait_for_ready_threaded(env_waiters, timeout):
        return False

    for waiter in env_waiters:
        for key in waiter.observed_resources:
            already_waited_on.add(key)

    end = time.time()
    elapsed = end - start
    timeout = int(timeout - elapsed)

    # wait on all ClowdApps in this namespace
    start = time.time()

    waiters = []
    for clowdapp in clowdapps["items"]:
        waiter = ResourceWaiter(
            namespace, "clowdapp", clowdapp["metadata"]["name"], watch_owned=True, watcher=watcher
        )
        waiters.append(waiter)
    if not wait_for_ready_threaded(waiters, timeout):
        return False

    for waiter in waiters:
        for key in waiter.observed_resources:
            already_waited_on.add(key)

    end = time.time()
    elapsed = end - start
    timeout = int(timeout - elapsed)

    # wait on anything else not covered by the above
    waiters = []
    for k, r in watcher.resources.copy().items():
        if r.restype in _resources_for_ns_wait() and r.key not in already_waited_on:
            waiter = ResourceWaiter(
                r.namespace,
                r.restype,
                r.name,
                watch_owned=True,
                watcher=watcher,
            )
            waiters.append(waiter)

    return wait_for_ready_threaded(waiters, timeout)


def wait_for_all_resources(namespace, timeout=600):
    # wrap the other wait_fors in 1 wait_for so overall timeout is honored
    # wait_for returns a tuple of the return code and the time taken
    watcher = ResourceWatcher(namespace)
    watcher.start()

    try:
        wait_for(
            _all_resources_ready,
            func_args=(namespace, timeout, watcher),
            message="wait for all deployed resources to be ready",
            timeout=timeout,
        )
    finally:
        watcher.stop()


def wait_for_db_resources(namespace, timeout=600):
    clowdapps = get_json("clowdapp", namespace=namespace).get("items", [])
    if len(clowdapps) == 0:
        raise ValueError(f"no clowdapps found in ns '{namespace}', no DB's to wait for")

    waiters = []
    for clowdapp in clowdapps:
        clowdapp_name = clowdapp["metadata"]["name"]
        db_name = clowdapp["spec"].get("database", {}).get("name")
        if db_name:
            waiters.append(ResourceWaiter(namespace, "deployment", f"{clowdapp_name}-db"))
        shared_db_app_name = clowdapp["spec"].get("database", {}).get("sharedDbAppName")
        if shared_db_app_name:
            waiters.append(ResourceWaiter(namespace, "deployment", f"{shared_db_app_name}-db"))

    if not waiters:
        raise ValueError(
            f"no clowdapps with db configurations found in '{namespace}', no DB's to wait for"
        )

    wait_for_ready_threaded(waiters, timeout)


def find_clowd_env_for_ns(ns):
    try:
        clowd_envs = get_json("clowdenvironment")
    except ValueError as err:
        if "unknown resource type" in str(err).lower():
            log.error("hit unknown resource type error, possibly not logged in to cluster?")
            clowd_envs = {"items": []}
        else:
            raise
    except ErrorReturnCode as err:
        log.debug("hit error running 'oc get clowdenvironment': %s", err)
        clowd_envs = {"items": []}

    for clowd_env in clowd_envs["items"]:
        target_ns = clowd_env["spec"].get("targetNamespace")
        # in case target ns was not defined in the spec, check the env's status...
        target_ns = target_ns or clowd_env.get("status", {}).get("targetNamespace")
        if target_ns == ns:
            return clowd_env


def get_clowd_env_target_ns(clowd_env_name):
    try:
        clowd_env = get_json("clowdenvironment", clowd_env_name)
    except ErrorReturnCode as err:
        log.debug("hit error running 'oc get clowdenvironment %s': %s", clowd_env_name, err)
        return None

    return clowd_env.get("status", {}).get("targetNamespace")


def wait_for_clowd_env_target_ns(clowd_env_name):
    log.info("waiting for Clowder to provision target namespace for env '%s'", clowd_env_name)
    return wait_for(
        get_clowd_env_target_ns,
        func_args=(clowd_env_name,),
        fail_condition=None,
        num_sec=60,
        message="wait for Clowder to provision target namespace",
    ).out


def wait_on_cji(namespace, cji_name, timeout):
    # wait for job associated with this CJI to appear
    log.info("waiting for Job to appear owned by CJI '%s'", cji_name)

    def _find_job():
        jobs = get_json("job", label=f"clowdjob={cji_name}", namespace=namespace)
        try:
            return jobs["items"][0]["metadata"]["name"]
        except (KeyError, IndexError):
            return False

    cji = Resource("clowdjobinvocation", cji_name, namespace)
    try:
        job_name, elapsed = wait_for(
            _find_job, num_sec=timeout, message=f"wait for Job to appear owned by CJI '{cji_name}'"
        )
    except TimedOutError:
        if not cji.ready:
            log.error("[%s] not ready, details: %s\n", cji.key, cji.details_str)
        raise

    log.info(
        "found Job '%s' created by CJI '%s', now waiting for pod to appear", job_name, cji_name
    )

    def _pod_found():
        pods = get_json("pod", label=f"job-name={job_name}", namespace=namespace)
        try:
            return pods["items"][0]["metadata"]["name"]
        except (KeyError, IndexError):
            return False

    remaining_time = timeout - elapsed

    pod_name, elapsed = wait_for(
        _pod_found,
        num_sec=remaining_time,
        message=f"wait for Pod to appear owned by CJI '{cji_name}'",
    )

    log.info(
        "found pod '%s' associated with CJI '%s', now waiting for pod to be 'running'",
        pod_name,
        cji_name,
    )

    remaining_time = remaining_time - elapsed

    waiter = ResourceWaiter(namespace, "pod", pod_name)
    waiter.wait_for_ready(remaining_time, reraise=True)

    return pod_name


def wait_on_reservation(res_name, timeout):
    log.info("waiting for reservation '%s' to get picked up by operator", res_name)

    def _find_reservation():
        res = get_json("reservation", name=res_name)
        return res.get("status", {}).get("namespace", False) or False

    ns_name, elapsed = wait_for(
        _find_reservation,
        num_sec=timeout,
        message=f"waiting for namespace to be allocated to reservation '{res_name}'",
    )
    return ns_name


def get_all_reservations():
    if not has_ns_operator():
        return []
    return get_json("reservation").get("items", [])


def check_for_existing_reservation(requester):
    if not has_ns_operator():
        return False

    log.info("Checking for existing reservations for '%s'", requester)

    for res in get_all_reservations():
        res_state = res.get("status", {}).get("state")
        if res["spec"]["requester"] == requester and res_state == "active":
            ns = res["status"]["namespace"]
            if get_json("namespace", ns):
                return True
            else:
                log.info("reservation found for namespace '%s' which no longer exists", ns)
    return False


def get_reservation(name=None, namespace=None, requester=None):
    if not has_ns_operator():
        return None

    if name:
        res = get_json("reservation", name=name)
        return res if res else False
    elif namespace:
        for res in get_all_reservations():
            if res.get("status", {}).get("namespace") == namespace:
                return res
    elif requester:
        requester_res = get_json("reservation", label=f"requester={requester}")
        numRes = len(requester_res.get("items", []))
        if numRes == 0:
            return None
        elif numRes == 1:
            return requester_res["items"][0]
        else:
            log.info("Multiple reservations found for requester '%s'. Aborting.", requester)
            return None

    return None


def get_pool_size_limit(pool):
    pool_data = get_json("namespacepool", name=pool)
    size_limit = pool_data["spec"].get("sizeLimit") if pool_data else 0
    return int(size_limit) if size_limit else 0


def get_reserved_namespace_quantity(pool):
    label = f"pool={pool}"
    pool_namespaces = get_all_namespaces(label=label)
    reserved_namespaces = [
        ns
        for ns in pool_namespaces
        if ns["metadata"].get("annotations", {}).get("reserved") == "true"
    ]

    return len(reserved_namespaces)



import pytest

from bonfire.utils import (
    get_version,
    hms_to_seconds,
    split_equals,
    validate_time_string,
    check_pypi,
    PYPI_URL,
)
from bonfire.utils import check_url_connection, FatalError


@pytest.mark.parametrize(
    "list_of_str, expected",
    [
        (["t1=test1", "t2=test2"], {"t1": "test1", "t2": "test2"}),
        (["t3=test3", "t4=test4"], {"t3": "test3", "t4": "test4"}),
    ],
)
def test_split_equals_pass(list_of_str: str, expected: str):
    result = split_equals(list_of_str)

    assert result == expected


@pytest.mark.parametrize(
    "list_of_str",
    [
        (["t1 = test1", "t2 = test2"],),
        (["t1 test1", "t2 test2"],),
    ],
)
def test_split_equals_raises_error(list_of_str: list):
    with pytest.raises(ValueError):
        split_equals(list_of_str)


@pytest.mark.parametrize(
    "time, expected",
    [
        ("30m", "30m"),
        ("1h45m", "1h45m"),
        ("1h20s", "1h20s"),
        ("2h15m32s", "2h15m32s"),
    ],
)
def test_validate_time_string(time: str, expected: str):
    result = validate_time_string(time)

    assert result == expected


@pytest.mark.parametrize("time", [("5m"), ("480h"), ("130")])
def test_validate_time_raises_error(time: str):
    with pytest.raises(ValueError):
        validate_time_string(time)


def test_get_version():
    result = get_version()

    assert result != "0.0.0"


@pytest.mark.parametrize(
    "seconds, expected",
    [
        ("1h", 3600),
        ("45m", 2700),
        ("65s", 65),
        ("1h30m", 5400),
    ],
)
def test_hms_to_seconds(seconds: str, expected: int):
    result = hms_to_seconds(seconds)

    assert result == expected


def test_url_connection_raises_on_invalid_url():
    with pytest.raises(ValueError, match=r".*invalidhosturl.*"):
        check_url_connection("foo-invalidhosturl")


def test_url_connection_checks_hostname(mocker):
    socket_library_mock = mocker.patch("bonfire.utils.socket.socket")
    socket_mock = socket_library_mock.return_value.__enter__.return_value
    check_url_connection("https://validhost.com")
    socket_mock.connect.assert_called_with(("validhost.com", 443))
    socket_mock.settimeout.assert_called_once()


def test_url_connection_timeout_handling(mocker):
    socket_library_mock = mocker.patch("bonfire.utils.socket.socket")
    socket_mock = socket_library_mock.return_value.__enter__.return_value
    socket_mock.connect.side_effect = TimeoutError("timed out!")

    with pytest.raises(FatalError, match=r"Unable to connect to.*after.*seconds.*is VPN needed.*"):
        check_url_connection("https://timingout.com")


def test_ip_timeout():
    with pytest.raises(FatalError, match="Unable to connect to.*after 1 seconds.*is VPN needed.*"):
        check_url_connection("https://10.255.255.1", timeout=1)


def test_url_connection_dns_lookup_fails():
    with pytest.raises(FatalError, match=r".*DNS lookup failed.*"):
        check_url_connection("https://baddomain.invalid")


def test_check_version_when_there_is_a_new_update(requests_mock, mocker):
    mocker.patch("bonfire.utils.Path").exists.return_value = False
    log_mock = mocker.patch("bonfire.utils.log")
    mocker.patch("bonfire.utils.importlib_metadata").version.return_value = "1.0.0"
    requests_mock.get(PYPI_URL, json={"info": {"version": "1.0.1"}})

    check_pypi()

    log_mock.info.assert_called()
    assert "checking pypi" in log_mock.info.call_args_list[0].args[0]
    assert "new release found" in log_mock.info.call_args_list[1].args[0]
    assert "yours: 1.0.0, available: 1.0.1" in log_mock.info.call_args_list[1].args[0]


def test_check_version_when_there_are_no_updates(requests_mock, mocker):
    mocker.patch("bonfire.utils.Path").exists.return_value = True
    log_mock = mocker.patch("bonfire.utils.log")
    mocker.patch("bonfire.utils.importlib_metadata").version.return_value = "1.0.0"
    requests_mock.get(PYPI_URL, json={"info": {"version": "1.0.0"}})

    check_pypi()

    log_mock.info.assert_called()
    assert "checking pypi" in log_mock.info.call_args_list[0].args[0]
    assert "new release found" not in log_mock.info.call_args_list[1].args[0]
    assert "up-to-date" in log_mock.info.call_args_list[1].args[0]


def test_check_version_unexpected_json(requests_mock, mocker):
    mocker.patch("bonfire.utils.Path").exists.return_value = False
    log_mock = mocker.patch("bonfire.utils.log")
    mocker.patch("bonfire.utils.importlib_metadata").version.return_value = "1.0.0"
    requests_mock.get(PYPI_URL, json={"woops": "not found"})

    check_pypi()

    log_mock.error.assert_called_once()
    assert "unable to parse version info" in log_mock.error.call_args_list[0].args[0]


def test_check_version_invalid_response(requests_mock, mocker):
    mocker.patch("bonfire.utils.Path").exists.return_value = False
    log_mock = mocker.patch("bonfire.utils.log")
    mocker.patch("bonfire.utils.importlib_metadata").version.return_value = "1.0.0"
    requests_mock.get(PYPI_URL, text="nothing!")

    check_pypi()

    log_mock.error.assert_called_once()
    assert "error fetching version from pypi" in log_mock.error.call_args_list[0].args[0]



# Alternative to 'sed' to search/replace within a file using regex

import os
import re
import sys

import click


def _do_lines(search_regex, replace_regex, file_path):
    new_lines = []
    with open(file_path, "r") as fp:
        for line in fp:
            new_lines.append(re.sub(search_regex, replace_regex, line))
    return "".join(new_lines)  # \n already included in the re.sub return data


def _do_file(search_regex, replace_regex, file_path):
    with open(file_path, "r") as fp:
        data = fp.read()
    return re.sub(search_regex, replace_regex, data, flags=re.DOTALL)


def _error(msg):
    click.error(msg)
    sys.exit(1)


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.argument("search_regex", required=True, type=str)
@click.argument("replace_regex", required=True, type=str)
@click.argument("file_path", required=True, type=str)
@click.option(
    "--in-place",
    "-i",
    is_flag=True,
    default=False,
    help="Overwrite file once replace is completed",
)
@click.option(
    "--lines",
    "-l",
    is_flag=True,
    default=False,
    help="Find re match on each line, not the entire file at once",
)
def main(search_regex, replace_regex, file_path, in_place, lines):
    if not os.path.exists(file_path):
        _error(f"file does not exist: {file_path}")
    if not os.path.isfile(file_path):
        _error(f"path is not a file: {file_path}")
    if not os.access(file_path, os.R_OK):
        _error(f"unable to file for reading: {file_path}")

    if lines:
        new_data = _do_lines(search_regex, replace_regex, file_path)
    else:
        new_data = _do_file(search_regex, replace_regex, file_path)

    if in_place:
        if not os.access(file_path, os.W_OK):
            _error(f"unable to file for writing: {file_path}")
        with open(file_path, "w") as fp:
            fp.write(new_data)
    else:
        print(new_data)


if __name__ == "__main__":
    main()



import uuid

import pytest

from bonfire.processor import TemplateProcessor, _should_remove
from bonfire.utils import RepoFile, AppOrComponentSelector


class MockRepoFile:
    """
    mock of utils.RepoFile so that we do not literally fetch templates from github/gitlab/etc.
    """

    templates = {}

    def __init__(self, name):
        self.name = name

    @classmethod
    def from_config(cls, data):
        return cls(data["name"])

    @classmethod
    def add_template(cls, name, template_commit, template_data):
        cls.templates[name] = {"commit": template_commit, "data": template_data}

    def fetch(self):
        return self.templates[self.name]["commit"], self.templates[self.name]["data"]


@pytest.fixture
def mock_repo_file(monkeypatch):
    MockRepoFile.templates = {}
    with monkeypatch.context() as m:
        m.setattr(RepoFile, "fetch", MockRepoFile.fetch)
        m.setattr(RepoFile, "from_config", MockRepoFile.from_config)
        yield MockRepoFile


SIMPLE_CLOWDAPP = """
---
apiVersion: v1
kind: Template
metadata:
  name: {name}-template
objects:
- apiVersion: cloud.redhat.com/v1alpha1
  kind: ClowdApp
  metadata:
    name: {name}
  spec:
    envName: ${{ENV_NAME}}
    dependencies: {deps}
    optionalDependencies: {optional_deps}
    deployments:
    - name: deployment1
      podSpec:
        resources:
          limits:
            cpu: ${{CPU_LIMIT_DEPLOYMENT1}}
            memory: ${{MEM_LIMIT_DEPLOYMENT1}}
          requests:
            cpu: ${{CPU_REQUEST_DEPLOYMENT1}}
            memory: ${{MEM_REQUEST_DEPLOYMENT1}}
    - name: deployment2
      podSpec:
        resources:
          limits:
            cpu: ${{DEPLOYMENT2_WRONG_NAME_CPU}}
            memory: ${{MEM_LIMIT_DEPLOYMENT2}}
          requests:
            cpu: ${{DEPLOYMENT2_WRONG_NAME_CPU}}
            memory: ${{MEMORY_REQUEST_DEPLOYMENT2}}
parameters:
- description: Image tag
  name: IMAGE_TAG
  required: true
- description: ClowdEnv Name
  name: ENV_NAME
  required: true
- name: CPU_LIMIT_DEPLOYMENT1
  value: 100m
- name: CPU_REQUEST_DEPLOYMENT1
  value: 1m
- name: MEM_REQUEST_DEPLOYMENT1
  value: 1Mi
- name: MEM_LIMIT_DEPLOYMENT1
  value: 100Mi
- name: DEPLOYMENT2_WRONG_NAME_CPU
  value: 2m
- name: MEMORY_REQUEST_DEPLOYMENT2
  value: 2Mi
- name: MEM_LIMIT_DEPLOYMENT2
  value: 200Mi
"""


def assert_clowdapps(items, app_list):
    found_apps = AppOrComponentSelector().apps
    for i in items:
        if i["kind"].lower() == "clowdapp":
            name = i["metadata"]["name"]
            found_apps.append(name)

    for app in app_list:
        if app not in found_apps:
            raise AssertionError(f"app {app} missing from processed output")

    for app in found_apps:
        if app not in app_list:
            raise AssertionError(f"app {app} should not be present in processed output")

    if len(set(found_apps)) != len(found_apps):
        raise AssertionError("apps present more than once in processed output")


def add_template(mock_repo_file, template_name, deps=None, optional_deps=None):
    deps = deps or []
    optional_deps = optional_deps or []
    mock_repo_file.add_template(
        template_name,
        uuid.uuid4().hex[0:6],
        SIMPLE_CLOWDAPP.format(name=template_name, deps=deps, optional_deps=optional_deps),
    )


def get_processor(apps_config):
    return TemplateProcessor(
        apps_config=apps_config,
        app_names=[],
        get_dependencies=True,
        optional_deps_method="hybrid",
        image_tag_overrides={},
        template_ref_overrides={},
        param_overrides={},
        clowd_env="some_env",
        remove_resources=AppOrComponentSelector(True, [], []),
        no_remove_resources=AppOrComponentSelector(False, [], []),
        remove_dependencies=AppOrComponentSelector(False, [], []),
        no_remove_dependencies=AppOrComponentSelector(True, [], []),
        single_replicas=True,
        component_filter=[],
        local=True,
        frontends=False,
    )


def get_apps_config():
    return {
        "app1": {
            "name": "app1",
            "components": [
                {
                    "name": "app1-component1",
                    "host": "local",
                    "repo": "test",
                    "path": "test",
                },
                {
                    "name": "app1-component2",
                    "host": "local",
                    "repo": "test",
                    "path": "test",
                },
            ],
        },
        "app2": {
            "name": "app2",
            "components": [
                {
                    "name": "app2-component1",
                    "host": "local",
                    "repo": "test",
                    "path": "test",
                },
                {
                    "name": "app2-component2",
                    "host": "local",
                    "repo": "test",
                    "path": "test",
                },
            ],
        },
        "app3": {
            "name": "app3",
            "components": [
                {
                    "name": "app3-component1",
                    "host": "local",
                    "repo": "test",
                    "path": "test",
                },
                {
                    "name": "app3-component2",
                    "host": "local",
                    "repo": "test",
                    "path": "test",
                },
            ],
        },
        "app4": {
            "name": "app4",
            "components": [
                {
                    "name": "app4-component1",
                    "host": "local",
                    "repo": "test",
                    "path": "test",
                },
                {
                    "name": "app4-component2",
                    "host": "local",
                    "repo": "test",
                    "path": "test",
                },
            ],
        },
    }


@pytest.mark.parametrize(
    "optional_deps_method,expected",
    [
        (
            "all",
            [
                "app1-component1",
                "app1-component2",
                "app2-component2",
                "app3-component2",
            ],
        ),
        (
            "hybrid",
            [
                "app1-component1",
                "app1-component2",
                "app2-component2",
                "app3-component2",
            ],
        ),
        (
            "none",
            [
                "app1-component1",
                "app1-component2",
                "app2-component2",
                "app3-component2",
            ],
        ),
    ],
)
def test_required_deps(mock_repo_file, optional_deps_method, expected):
    """
    app1-component1 has 'app2-component2' listed under 'dependencies'
    app2-component2 has 'app3-component2' listed under 'dependencies'

    test that processing app1 results in expected ClowdApp dependencies being pulled in
    """
    add_template(mock_repo_file, "app1-component1", deps=["app2-component2"])
    add_template(mock_repo_file, "app1-component2")
    add_template(mock_repo_file, "app2-component2", deps=["app3-component2"])
    # template for app3-component2 will contain a dep we've already handled
    add_template(mock_repo_file, "app3-component2", deps=["app1-component1"])

    processor = get_processor(get_apps_config())
    processor.optional_deps_method = optional_deps_method
    processor.requested_app_names = ["app1"]
    processed = processor.process()
    assert_clowdapps(processed["items"], expected)


@pytest.mark.parametrize(
    "optional_deps_method,expected",
    [
        (
            "all",
            [
                "app1-component1",
                "app1-component2",
                "app2-component2",
                "app3-component2",
            ],
        ),
        ("hybrid", ["app1-component1", "app1-component2", "app2-component2"]),
        ("none", ["app1-component1", "app1-component2"]),
    ],
)
def test_optional_deps(mock_repo_file, optional_deps_method, expected):
    """
    app1-component1 has 'app2-component2' listed under 'optionalDependencies'
    app2-component2 has 'app3-component2' listed under 'optionalDependencies'

    test that processing app1 results in expected ClowdApp dependencies being pulled in depending
    on what 'optional dependencies mode' is selected
    """
    add_template(mock_repo_file, "app1-component1", optional_deps=["app2-component2"])
    add_template(mock_repo_file, "app1-component2")
    add_template(mock_repo_file, "app2-component2", optional_deps=["app3-component2"])
    # template for app3-component2 will contain a dep we've already handled
    add_template(mock_repo_file, "app3-component2", deps=["app1-component1"])

    processor = get_processor(get_apps_config())
    processor.optional_deps_method = optional_deps_method
    processor.requested_app_names = ["app1"]
    processed = processor.process()
    assert_clowdapps(processed["items"], expected)


@pytest.mark.parametrize(
    "optional_deps_method,expected",
    [
        (
            "all",
            [
                "app1-component1",
                "app1-component2",
                "app2-component1",
                "app2-component2",
                "app3-component1",
                "app3-component2",
            ],
        ),
        (
            "hybrid",
            [
                "app1-component1",
                "app1-component2",
                "app2-component1",
                "app3-component1",
                "app3-component2",
            ],
        ),
        ("none", ["app1-component1", "app1-component2", "app3-component1"]),
    ],
)
def test_mixed_deps(mock_repo_file, optional_deps_method, expected):
    """
    app1-component1 has 'app3-component1' listed under 'dependencies'
    app1-component1 has 'app2-component1' listed under 'optionalDependencies'

    app2-component1 has 'app3-component2' listed under 'dependencies'
    app2-component1 has 'app2-component2' listed under 'optionalDependencies'

    test that processing app1 results in expected ClowdApp dependencies being pulled in depending
    on what 'optional dependencies mode' is selected
    """
    add_template(
        mock_repo_file,
        "app1-component1",
        deps=["app3-component1"],
        optional_deps=["app2-component1"],
    )
    add_template(
        mock_repo_file,
        "app2-component1",
        deps=["app3-component2"],
        optional_deps=["app2-component2"],
    )
    add_template(mock_repo_file, "app1-component2")
    add_template(mock_repo_file, "app2-component2")
    add_template(mock_repo_file, "app3-component1")
    add_template(mock_repo_file, "app3-component2")

    processor = get_processor(get_apps_config())
    processor.optional_deps_method = optional_deps_method
    processor.requested_app_names = ["app1"]
    processed = processor.process()
    assert_clowdapps(processed["items"], expected)


@pytest.mark.parametrize(
    "optional_deps_method,expected",
    [
        (
            "all",
            [
                "app1-component1",
                "app1-component2",
                "app2-component1",
                "app2-component2",
                "app3-component1",
                "app3-component2",
                "app4-component1",
                "app4-component2",
            ],
        ),
        (
            "hybrid",
            [
                "app1-component1",
                "app1-component2",
                "app2-component1",
                "app2-component2",
                "app3-component1",
                "app3-component2",
            ],
        ),
        (
            "none",
            [
                "app1-component1",
                "app1-component2",
                "app2-component1",
                "app3-component1",
                "app3-component2",
            ],
        ),
    ],
)
def test_mixed_deps_two_apps(mock_repo_file, optional_deps_method, expected):
    """
    app1-component1 has 'app2-component1' listed under 'dependencies'
    app1-component1 has 'app3-component1' listed under 'optionalDependencies'

    app2-component1 has 'app4-component1' listed under 'optionalDependencies'
    app2-component2 has 'app4-component2' listed under 'optionalDependencies'

    app3-commponent1 has 'app2-component2' listed under 'optionalDependencies'

    test that processing app1 and app3 results in expected ClowdApp dependencies being pulled in
    depending on what 'optional dependencies mode' is selected
    """
    add_template(
        mock_repo_file,
        "app1-component1",
        deps=["app2-component1"],
        optional_deps=["app3-component1"],
    )
    add_template(mock_repo_file, "app1-component2")
    add_template(mock_repo_file, "app2-component1", optional_deps=["app4-component1"])
    add_template(mock_repo_file, "app2-component2", optional_deps=["app4-component2"])
    add_template(mock_repo_file, "app3-component1", optional_deps=["app2-component2"])
    add_template(mock_repo_file, "app3-component2")
    add_template(mock_repo_file, "app4-component1")
    add_template(mock_repo_file, "app4-component2")

    processor = get_processor(get_apps_config())
    processor.optional_deps_method = optional_deps_method
    processor.requested_app_names = ["app1", "app3"]
    processed = processor.process()
    assert_clowdapps(processed["items"], expected)


# Testing --no-remove-resources/dependency "app:" syntax
def test_should_remove_remove_for_none_no_exceptions():
    # --no-remove-resources all --no-remove-resources component1 --no-remove-resources app:app1
    remove_resources = AppOrComponentSelector(select_all=False, components=[], apps=[])
    no_remove_resources = AppOrComponentSelector(
        select_all=True, components=["component1"], apps=["app1"]
    )

    assert _should_remove(remove_resources, no_remove_resources, "app2", "component2") is False
    assert _should_remove(remove_resources, no_remove_resources, "app2", "component1") is False
    assert _should_remove(remove_resources, no_remove_resources, "app1", "whatever") is False


def test_should_remove_remove_for_all_no_exceptions():
    # --remove-resources all --remove-resources component1 --remove-resources app:app1
    remove_resources = AppOrComponentSelector(
        select_all=True, components=["component1"], apps=["app1"]
    )
    no_remove_resources = AppOrComponentSelector(select_all=False, components=[], apps=[])

    assert _should_remove(remove_resources, no_remove_resources, "app2", "component2") is True
    assert _should_remove(remove_resources, no_remove_resources, "app2", "component1") is True
    assert _should_remove(remove_resources, no_remove_resources, "app1", "whatever") is True


def test_should_remove_remove_option_select_all():
    # --remove-resources all --no-remove-resources component1
    assert (
        _should_remove(
            AppOrComponentSelector(select_all=True, components=[], apps=[]),
            AppOrComponentSelector(select_all=False, components=["component1"], apps=[]),
            "app2",
            "component1",
        )
        is False
    )

    # --remove-resources all --no-remove-resources app:app1
    assert (
        _should_remove(
            AppOrComponentSelector(select_all=True, components=[], apps=[]),
            AppOrComponentSelector(select_all=False, components=[], apps=["app1"]),
            "app1",
            "component1",
        )
        is False
    )

    # --remove-resources all
    assert (
        _should_remove(
            AppOrComponentSelector(select_all=True, components=[], apps=[]),
            AppOrComponentSelector(select_all=False, components=[], apps=[]),
            "app1",
            "component2",
        )
        is True
    )


def test_should_remove_no_remove_option_select_all():
    # --remove-resources component1 --no-remove-resources all
    assert (
        _should_remove(
            AppOrComponentSelector(select_all=False, components=["component1"], apps=[]),
            AppOrComponentSelector(select_all=True, components=[], apps=[]),
            "app1",
            "component1",
        )
        is True
    )

    # --remove-resources app:app1 --no-remove-resources all
    assert (
        _should_remove(
            AppOrComponentSelector(select_all=False, components=[], apps=["app1"]),
            AppOrComponentSelector(select_all=True, components=[], apps=[]),
            "app1",
            "component1",
        )
        is True
    )

    # --remove-resources component1 --remove-resources app:app1 --no-remove-resources all
    assert (
        _should_remove(
            AppOrComponentSelector(select_all=False, components=["component1"], apps=["app1"]),
            AppOrComponentSelector(select_all=True, components=[], apps=[]),
            "app1",
            "component1",
        )
        is True
    )

    # --remove-resources component1 --remove-resources app:app2 --no-remove-resources all
    assert (
        _should_remove(
            AppOrComponentSelector(select_all=False, components=["component1"], apps=["app1"]),
            AppOrComponentSelector(select_all=True, components=[], apps=[]),
            "app2",
            "component1",
        )
        is True
    )

    # --remove-resources component2 --remove-resources app:app1 --no-remove-resources all
    assert (
        _should_remove(
            AppOrComponentSelector(select_all=False, components=["component1"], apps=["app1"]),
            AppOrComponentSelector(select_all=True, components=[], apps=[]),
            "app1",
            "component2",
        )
        is True
    )

    # --remove-resources component2 --remove-resources app:app2 --no-remove-resources all
    assert (
        _should_remove(
            AppOrComponentSelector(select_all=False, components=["component1"], apps=["app1"]),
            AppOrComponentSelector(select_all=True, components=[], apps=[]),
            "app2",
            "component2",
        )
        is False
    )

    assert (
        # --no-remove-resources all
        _should_remove(
            AppOrComponentSelector(select_all=False, components=[], apps=[]),
            AppOrComponentSelector(select_all=True, components=[], apps=[]),
            "app1",
            "component1",
        )
        is False
    )


@pytest.mark.parametrize("default", (True, False), ids=("default=True", "default=False"))
def test_should_remove_component_overrides_app(default):
    # --no-remove-resources app:app1 --remove-resources component1
    remove_resources = AppOrComponentSelector(select_all=False, components=["component2"], apps=[])
    no_remove_resources = AppOrComponentSelector(select_all=False, components=[], apps=["app1"])

    assert (
        _should_remove(remove_resources, no_remove_resources, "app1", "component1", default)
        is False
    )
    assert (
        _should_remove(remove_resources, no_remove_resources, "app1", "component2", default) is True
    )
    assert (
        _should_remove(remove_resources, no_remove_resources, "anything", "else", default)
        is default
    )


@pytest.mark.parametrize("default", (True, False), ids=("default=True", "default=False"))
def test_should_remove_component_app_combos(default):
    # --no-remove-resources app:app2 --no-remove-resources component2 \
    #   --remove-resources component1 --remove-resources app:app1
    remove_resources = AppOrComponentSelector(
        select_all=False, components=["component1"], apps=["app1"]
    )
    no_remove_resources = AppOrComponentSelector(
        select_all=False, components=["component2"], apps=["app2"]
    )

    assert (
        _should_remove(remove_resources, no_remove_resources, "app2", "component1", default) is True
    )
    assert (
        _should_remove(remove_resources, no_remove_resources, "app1", "anything", default) is True
    )
    assert (
        _should_remove(remove_resources, no_remove_resources, "app1", "component2", default)
        is False
    )
    assert (
        _should_remove(remove_resources, no_remove_resources, "app2", "anything", default) is False
    )
    assert (
        _should_remove(remove_resources, no_remove_resources, "anything", "else", default)
        is default
    )


def get_apps_config_with_params(parameters=None):
    return {
        "app1": {
            "name": "app1",
            "components": [
                {
                    "name": "app1-component1",
                    "host": "local",
                    "repo": "test",
                    "path": "test",
                    "parameters": parameters or {},
                },
            ],
        },
    }


def test_remove_resources_untrusted_params(mock_repo_file):
    """
    Test that resource configs are removed if template's parameter values are not explicitly set
    """
    add_template(mock_repo_file, "app1-component1")
    apps_config = get_apps_config_with_params(None)
    processor = get_processor(apps_config)
    processor.requested_app_names = ["app1"]
    result = processor.process()

    deployments = result["items"][0]["spec"]["deployments"]
    deployment1, deployment2 = deployments[0], deployments[1]

    assert deployment1["podSpec"]["resources"]["requests"] == {}
    assert deployment1["podSpec"]["resources"]["limits"] == {}
    assert deployment2["podSpec"]["resources"]["requests"] == {}
    assert deployment2["podSpec"]["resources"]["limits"] == {}


def test_preserve_resources_trusted_params(mock_repo_file):
    """
    Test that using trusted parameters causes cpu/mem configurations to be preserved.

    Ensures that a value set with an untrusted parameter name is still removed.
    """
    add_template(mock_repo_file, "app1-component1")
    apps_config = get_apps_config_with_params(
        parameters={
            "CPU_LIMIT_DEPLOYMENT1": "456m",
            "CPU_REQUEST_DEPLOYMENT1": "123m",
            "MEM_LIMIT_DEPLOYMENT1": "456Mi",
            "MEM_REQUEST_DEPLOYMENT1": "123Mi",
            "DEPLOYMENT2_WRONG_NAME_CPU": "1",
            "MEM_LIMIT_DEPLOYMENT2": "910Mi",
            "MEMORY_REQUEST_DEPLOYMENT2": "789Mi",
        }
    )
    processor = get_processor(apps_config)
    processor.requested_app_names = ["app1"]
    result = processor.process()

    deployments = result["items"][0]["spec"]["deployments"]
    deployment1, deployment2 = deployments[0], deployments[1]

    assert deployment1["podSpec"]["resources"]["requests"]["cpu"] == "123m"
    assert deployment1["podSpec"]["resources"]["requests"]["memory"] == "123Mi"
    assert deployment1["podSpec"]["resources"]["limits"]["cpu"] == "456m"
    assert deployment1["podSpec"]["resources"]["limits"]["memory"] == "456Mi"
    assert deployment2["podSpec"]["resources"]["requests"]["memory"] == "789Mi"
    assert deployment2["podSpec"]["resources"]["limits"]["memory"] == "910Mi"
    # deployment2 CPU param does not match trusted syntax for name
    assert "cpu" not in deployment2["podSpec"]["resources"]["requests"]
    assert "cpu" not in deployment2["podSpec"]["resources"]["limits"]


@pytest.mark.parametrize(
    "no_remove_resources",
    (
        # --no-remove-resources all
        AppOrComponentSelector(select_all=True, components=[], apps=[]),
        # --no-remove-resources app:app1
        AppOrComponentSelector(select_all=False, components=[], apps=["app1"]),
        # --no-remove-resources app1-component1
        AppOrComponentSelector(select_all=False, components=["app1-component1"], apps=[]),
    ),
    ids=("all", "app", "component"),
)
def test_preserve_resources_cli_option(mock_repo_file, no_remove_resources):
    """
    Test that using "--no-remove-resources" causes cpu/mem configs to be preserved
    """
    add_template(mock_repo_file, "app1-component1")
    apps_config = get_apps_config_with_params(parameters=None)
    processor = get_processor(apps_config)
    processor.requested_app_names = ["app1"]
    processor.no_remove_resources = no_remove_resources
    processor.remove_resources = AppOrComponentSelector(False, [], [])
    result = processor.process()

    deployments = result["items"][0]["spec"]["deployments"]
    deployment1, deployment2 = deployments[0], deployments[1]

    assert deployment1["podSpec"]["resources"]["requests"]["cpu"] == "1m"
    assert deployment1["podSpec"]["resources"]["requests"]["memory"] == "1Mi"
    assert deployment1["podSpec"]["resources"]["limits"]["cpu"] == "100m"
    assert deployment1["podSpec"]["resources"]["limits"]["memory"] == "100Mi"
    assert deployment2["podSpec"]["resources"]["requests"]["memory"] == "2Mi"
    assert deployment2["podSpec"]["resources"]["requests"]["cpu"] == "2m"
    assert deployment2["podSpec"]["resources"]["limits"]["memory"] == "200Mi"
    assert deployment2["podSpec"]["resources"]["limits"]["cpu"] == "2m"



from datetime import datetime as dt
import logging
import json
import requests
import sys
import uuid
from concurrent.futures import ThreadPoolExecutor

import bonfire.config as conf


log = logging.getLogger(__name__)


class ElasticLogger:
    def __init__(self):
        self.es_telemetry = logging.getLogger("elasicsearch")

        # prevent duplicate handlers
        self.es_handler = next(
            (h for h in self.es_telemetry.handlers if type(h) is AsyncElasticsearchHandler), None
        )
        if not self.es_handler:
            self.es_handler = AsyncElasticsearchHandler(
                f"{conf.ELASTICSEARCH_HOST}/{conf.ELASTICSEARCH_INDEX}/_doc/"
            )
            self.es_telemetry.addHandler(self.es_handler)

    def send_telemetry(self, log_message, success=True):
        self.es_handler.set_success_status(success)

        self.es_telemetry.info(log_message)


class AsyncElasticsearchHandler(logging.Handler):
    def __init__(self, es_url):
        super().__init__()
        self.es_url = es_url
        self.executor = ThreadPoolExecutor(max_workers=10)
        self.start_time = dt.now()
        self.metadata = {
            "uuid": str(uuid.uuid4()),
            "start_time": self.start_time.isoformat(),
            "bot": conf.BONFIRE_BOT,
            "client_id": conf.CLIENT_ID,
            "command": self._mask_parameter_values(sys.argv[1:]),
        }

    def emit(self, record):
        self.metadata["@timestamp"] = dt.now().isoformat()
        self.metadata["elapsed_sec"] = (dt.now() - self.start_time).total_seconds()

        log_entry = {"log": self.format(record), "metadata": self.metadata}

        if conf.ENABLE_TELEMETRY:
            if not conf.ELASTICSEARCH_APIKEY or not conf.ELASTICSEARCH_HOST:
                log.error("Bonfire telemetry secret(s) not set")
                return

            self.executor.submit(self.send_to_es, json.dumps(log_entry))

    def set_success_status(self, run_status):
        self.metadata["succeeded"] = run_status

    def send_to_es(self, log_entry):
        # Convert log_entry to JSON and send to Elasticsearch
        log.info("Sending telemetry data...")

        try:
            headers = {
                "Authorization": conf.ELASTICSEARCH_APIKEY,
                "Content-Type": "application/json",
            }

            response = requests.post(self.es_url, headers=headers, data=log_entry, timeout=0.1)
            response.raise_for_status()
            log.info("Successfully sent telemetry data")
        except Exception as e:
            # Handle exceptions (e.g., network issues, Elasticsearch down)
            log.error("Error sending data to elasticsearch: %s", e)

    @staticmethod
    def _mask_parameter_values(cli_args):
        masked_list = []

        is_parameter = False
        for arg in cli_args:
            if is_parameter:
                masked_arg = f"{arg.split('=')[0]}=*******"
                masked_list.append(masked_arg)
                is_parameter = False
            else:
                masked_list.append(arg)
                is_parameter = arg == "-p" or arg == "--set-parameter"

        return masked_list



import bonfire
from bonfire.qontract import get_apps_for_env, sub_refs, ENVS_QUERY, APPS_QUERY


def _mock_envs_gql_resp():
    return {
        "envs": [
            {
                "name": "env_with_no_apps",
            },
            {
                "name": "ephemeral",
                "parameters": '{"PARAM_1":"ephemeral1","PARAM_2":"ephemeral2","PARAM_3":100}',
                "namespaces": [
                    {"name": "ephemeral-base", "path": "/path/to/ephemeral-base.yml"},
                ],
            },
            {
                "name": "stage",
                "parameters": '{"PARAM_1":"stage1","PARAM_2":"stage2","PARAM_3":200}',
                "namespaces": [
                    {"name": "stage-namespace-1", "path": "/path/to/stage-namespace-1.yml"},
                    {"name": "stage-namespace-2", "path": "/path/to/stage-namespace-2.yml"},
                    {"name": "stage-namespace-3", "path": "/path/to/stage-namespace-3.yml"},
                ],
            },
            {
                "name": "prod",
                "parameters": '{"PARAM_1":"prod1","PARAM_2":"prod2","PARAM_3":300}',
                "namespaces": [
                    {"name": "prod-namespace-1", "path": "/path/to/prod-namespace-1.yml"},
                    {"name": "prod-namespace-2", "path": "/path/to/prod-namespace-2.yml"},
                ],
            },
        ]
    }


def _mock_apps_gql_resp():
    return {
        "apps": [
            {
                "name": "app1",
                "parentApp": {"name": "insights"},
                "saasFiles": [
                    {
                        "path": "/path/to/deploy.yml",
                        "name": "app1",
                        "parameters": None,
                        "resourceTemplates": [
                            {
                                "name": "component1",
                                "path": "/deploy/template.yml",
                                "url": "https://github.test/Org/Repo",
                                "parameters": None,
                                "targets": [
                                    {
                                        "namespace": {
                                            "name": "app1-stage-ns-1",
                                            "path": "/path/to/stage-namespace-1.yml",
                                            "cluster": {"name": "test_cluster"},
                                        },
                                        "ref": "master",
                                        "parameters": None,
                                    },
                                    {
                                        "namespace": {
                                            "name": "app1-stage-ns-2",
                                            "path": "/path/to/stage-namespace-2.yml",
                                            "cluster": {"name": "test_cluster"},
                                        },
                                        "ref": "abc1234",
                                        "parameters": '{"FAVORED_PARAM":"favored.value"}',
                                    },
                                    {
                                        "namespace": {
                                            "name": "app1-stage-ns-3",
                                            "path": "/path/to/stage-namespace-3.yml",
                                            "cluster": {"name": "test_cluster"},
                                        },
                                        "ref": "xyz6789",
                                        "parameters": '{"FAVORED_PARAM":"unfavored.value"}',
                                    },
                                    {
                                        "namespace": {
                                            "name": "app1-prod-ns-1",
                                            "path": "/path/to/prod-namespace-2.yml",
                                            "cluster": {"name": "test_cluster"},
                                        },
                                        "ref": "prod1ref",
                                        "parameters": '{"REPLICAS":1,"FAVORED_PARAM":"favored"}',
                                    },
                                    {
                                        "namespace": {
                                            "name": "app1-prod-ns-2",
                                            "path": "/path/to/prod-namespace-2.yml",
                                            "cluster": {"name": "test_cluster"},
                                        },
                                        "ref": "prod2ref",
                                        "parameters": '{"REPLICAS":0,"FAVORED_PARAM":"favored"}',
                                    },
                                    {
                                        "namespace": {
                                            "name": "ephemeral-base",
                                            "path": "/path/to/ephemeral-base.yml",
                                            "cluster": {"name": "test_cluster"},
                                        },
                                        "ref": "internal",
                                        "parameters": None,
                                    },
                                ],
                            },
                        ],
                    }
                ],
            },
        ]
    }


class MockGQLClient:
    def execute(self, query):
        if query == ENVS_QUERY:
            return _mock_envs_gql_resp()
        elif query == APPS_QUERY:
            return _mock_apps_gql_resp()
        else:
            raise ValueError("invalid query for MockGQLClient")


class MockAppInterfaceClient(bonfire.qontract.Client):
    def __init__(self):
        self.client = MockGQLClient()


def _mock_get_client():
    return MockAppInterfaceClient()


def test_no_pref(monkeypatch):
    """
    Test that with no preference, git ref from first stage target is chosen
    """
    monkeypatch.setattr(bonfire.qontract, "get_client", _mock_get_client)
    expected_apps = {
        "app1": {
            "name": "app1",
            "components": [
                {
                    "name": "component1",
                    "path": "/deploy/template.yml",
                    "host": "github",
                    "repo": "Org/Repo",
                    "ref": "master",
                    "parameters": {
                        "PARAM_1": "ephemeral1",
                        "PARAM_2": "ephemeral2",
                        "PARAM_3": 100,
                    },
                }
            ],
        }
    }
    ephemeral_apps = get_apps_for_env(env_name="ephemeral", preferred_params={})
    final_apps = sub_refs(ephemeral_apps, "stage", None, preferred_params={})

    assert final_apps == expected_apps


def test_preferred_ref(monkeypatch):
    """
    Test that git ref from stage target with FAVORED_PARAM=favored.value is chosen
    """
    monkeypatch.setattr(bonfire.qontract, "get_client", _mock_get_client)
    expected_apps = {
        "app1": {
            "name": "app1",
            "components": [
                {
                    "name": "component1",
                    "path": "/deploy/template.yml",
                    "host": "github",
                    "repo": "Org/Repo",
                    "ref": "abc1234",
                    "parameters": {
                        "PARAM_1": "ephemeral1",
                        "PARAM_2": "ephemeral2",
                        "PARAM_3": 100,
                    },
                }
            ],
        }
    }
    prefer = {"FAVORED_PARAM": "favored.value"}
    ephemeral_apps = get_apps_for_env(env_name="ephemeral", preferred_params=prefer)
    final_apps = sub_refs(ephemeral_apps, "stage", None, preferred_params=prefer)

    assert final_apps == expected_apps


def test_prefer_replicas(monkeypatch):
    """
    Test that git ref from prod target with REPLICAS=1 is chosen
    """
    monkeypatch.setattr(bonfire.qontract, "get_client", _mock_get_client)
    expected_apps = {
        "app1": {
            "name": "app1",
            "components": [
                {
                    "name": "component1",
                    "path": "/deploy/template.yml",
                    "host": "github",
                    "repo": "Org/Repo",
                    "ref": "prod1ref",
                    "parameters": {
                        "PARAM_1": "ephemeral1",
                        "PARAM_2": "ephemeral2",
                        "PARAM_3": 100,
                    },
                }
            ],
        }
    }
    prefer = {"FAVORED_PARAM": "favored"}
    ephemeral_apps = get_apps_for_env(env_name="ephemeral", preferred_params=prefer)
    final_apps = sub_refs(ephemeral_apps, "prod", None, preferred_params=prefer)

    assert final_apps == expected_apps


def test_fallback_with_preference(monkeypatch):
    """
    Test that git ref from stage target with FAVORED_PARAM=favored.value is chosen when
    ref env is set to 'prod' but no deploy config is present for the component in 'prod'
    """
    monkeypatch.setattr(bonfire.qontract, "get_client", _mock_get_client)
    expected_apps = {
        "app1": {
            "name": "app1",
            "components": [
                {
                    "name": "component1",
                    "path": "/deploy/template.yml",
                    "host": "github",
                    "repo": "Org/Repo",
                    "ref": "abc1234",
                    "parameters": {
                        "PARAM_1": "ephemeral1",
                        "PARAM_2": "ephemeral2",
                        "PARAM_3": 100,
                    },
                }
            ],
        }
    }
    prefer = {"FAVORED_PARAM": "favored.value"}
    ephemeral_apps = get_apps_for_env(env_name="ephemeral", preferred_params=prefer)
    final_apps = sub_refs(ephemeral_apps, "env_with_no_apps", "stage", preferred_params=prefer)

    assert final_apps == expected_apps



"""
Handles importing of configmaps from a local directory
"""

import glob
import json
import logging
import os

from ocviapy import get_json, oc

from bonfire.utils import FatalError, load_file

log = logging.getLogger(__name__)


def _parse_configmaps_file(path):
    """
    Return a dict of all configmaps in a file with key: configmap name,
    val: parsed configmap json/yaml
    The file can contain 1 configmap, or a list of configmaps
    """
    content = load_file(path)
    configmaps = {}
    if content.get("kind").lower() == "list":
        items = content.get("items", [])
    else:
        items = [content]

    for item in items:
        if item.get("kind").lower() == "configmap":
            try:
                configmaps[item["metadata"]["name"]] = item
            except KeyError:
                raise FatalError("Configmap at path '{}' has no metadata/name".format(path))

    return configmaps


def _get_files_in_dir(path):
    """
    Get a list of all .yml/.yaml/.json files in a dir
    """
    files = list(glob.glob(os.path.join(path, "*.yaml")))
    files.extend(list(glob.glob(os.path.join(path, "*.yml"))))
    files.extend(list(glob.glob(os.path.join(path, "*.json"))))
    return files


def _import_configmaps(configmap_name, configmap_data):
    # get existing configmap in the ns (if it exists)
    current_configmap = get_json("configmap", configmap_name) or {}

    # avoid race conditions when running multiple processes by comparing the data
    data_mismatch = current_configmap.get("data") != configmap_data.get("data")
    if data_mismatch:
        log.info("replacing configmap '%s' using local copy", configmap_name)
        # delete from dst ns so that applying 'null' values will work
        oc("delete", "--ignore-not-found", "configmap", configmap_name, _silent=True)
        oc("apply", "-f", "-", _silent=True, _in=json.dumps(configmap_data))


def import_configmaps_from_dir(path):
    if not os.path.exists(path):
        raise FatalError(f"configmaps directory not found: {path}")

    if not os.path.isdir(path):
        raise FatalError(f"invalid configmaps directory: {path}")

    files = _get_files_in_dir(path)
    configmaps = {}
    log.info("importing configmaps from local path: %s", path)
    for confmaps_file in files:
        confmaps_in_file = _parse_configmaps_file(confmaps_file)
        log.info(
            "loaded %d configmap(s) from file '%s'",
            len(confmaps_in_file),
            confmaps_file,
        )
        for confmaps_name in confmaps_in_file:
            if confmaps_name in configmaps:
                raise FatalError(
                    f"configmap with name '{confmaps_name}' defined twice in configmaps dir"
                )
        configmaps.update(confmaps_in_file)

    for configmap_name, configmap_data in configmaps.items():
        _import_configmaps(configmap_name, configmap_data)



import pytest

from bonfire.bonfire import _get_apps_config, APP_SRE_SRC, FILE_SRC

import bonfire

# Make sure to use functions for these test dictionaries instead of global vars.
# Otherwise, data gets polluted between the tests due to re-using the same dict object in memory.


def _target_apps():
    return {
        "appA": {
            "name": "appA",
            "components": [
                {
                    "name": "appAcomponent1",
                    "host": "github",
                    "repo": "someorg/appAcomponent1",
                    "path": "deploy/template.yaml",
                    "ref": "appAcomponent1_target_ref",
                    "parameters": {},
                },
                {
                    "name": "appAcomponent2",
                    "host": "github",
                    "repo": "someorg/appAcomponent2",
                    "path": "deploy/template.yaml",
                    "ref": "appAcomponent2_target_ref",
                    "parameters": {},
                },
            ],
        },
        "appB": {
            "name": "appB",
            "components": [
                {
                    "name": "appBcomponent1",
                    "host": "github",
                    "repo": "someorg/appBcomponent1",
                    "path": "deploy/template.yaml",
                    "ref": "appBcomponent1_target_ref",
                    "parameters": {},
                },
                {
                    "name": "appBcomponent2",
                    "host": "github",
                    "repo": "someorg/appBcomponent2",
                    "path": "deploy/template.yaml",
                    "ref": "appBcomponent2_target_ref",
                    "parameters": {
                        "EXISTING_PARAM1": "EXISTING_VALUE1",
                        "EXISTING_PARAM2": "EXISTING_VALUE2",
                    },
                },
            ],
        },
    }


def _reference_apps():
    return {
        "appA": {
            "name": "appA",
            "components": [
                {
                    "name": "appAcomponent1",
                    "host": "github",
                    "repo": "someorg/appAcomponent1",
                    "path": "deploy/template.yaml",
                    "ref": "appAcomponent1_reference_ref",
                    "parameters": {},
                },
                {
                    "name": "appAcomponent2",
                    "host": "github",
                    "repo": "someorg/appAcomponent2",
                    "path": "deploy/template.yaml",
                    "ref": "appAcomponent2_reference_ref",
                    "parameters": {},
                },
            ],
        },
        "appB": {
            "name": "appB",
            "components": [
                {
                    "name": "appBcomponent1",
                    "host": "github",
                    "repo": "someorg/appBcomponent1",
                    "path": "deploy/template.yaml",
                    "ref": "appBcomponent1_reference_ref",
                    "parameters": {},
                },
                {
                    "name": "appBcomponent2",
                    "host": "github",
                    "repo": "someorg/appBcomponent2",
                    "path": "deploy/template.yaml",
                    "ref": "appBcomponent2_reference_ref",
                    "parameters": {},
                },
            ],
        },
    }


def _target_apps_w_refs_subbed():
    return {
        "appA": {
            "name": "appA",
            "components": [
                {
                    "name": "appAcomponent1",
                    "host": "github",
                    "repo": "someorg/appAcomponent1",
                    "path": "deploy/template.yaml",
                    "ref": "appAcomponent1_reference_ref",
                    "parameters": {},
                },
                {
                    "name": "appAcomponent2",
                    "host": "github",
                    "repo": "someorg/appAcomponent2",
                    "path": "deploy/template.yaml",
                    "ref": "appAcomponent2_reference_ref",
                    "parameters": {},
                },
            ],
        },
        "appB": {
            "name": "appB",
            "components": [
                {
                    "name": "appBcomponent1",
                    "host": "github",
                    "repo": "someorg/appBcomponent1",
                    "path": "deploy/template.yaml",
                    "ref": "appBcomponent1_reference_ref",
                    "parameters": {},
                },
                {
                    "name": "appBcomponent2",
                    "host": "github",
                    "repo": "someorg/appBcomponent2",
                    "path": "deploy/template.yaml",
                    "ref": "appBcomponent2_reference_ref",
                    "parameters": {
                        "EXISTING_PARAM1": "EXISTING_VALUE1",
                        "EXISTING_PARAM2": "EXISTING_VALUE2",
                    },
                },
            ],
        },
    }


def _mock_get_apps_for_env(env, preferred_params):
    if env is None or env == "test_env_with_no_apps":
        return {}
    elif env == "test_target_env":
        return _target_apps()
    elif env == "test_ref_env":
        return _reference_apps()
    else:
        raise ValueError(f"invalid env '{env}' provided to mock function")


def _setup_monkeypatch(monkeypatch, source, local_cfg):
    # always patch this func since it is used in bonfire.qontract.sub_refs()
    monkeypatch.setattr(bonfire.qontract, "get_apps_for_env", _mock_get_apps_for_env)

    # patch loading of local cfg
    monkeypatch.setattr(bonfire.bonfire.conf, "load_config", lambda _: local_cfg)

    # patch fetching of remote app config
    if source == APP_SRE_SRC:
        monkeypatch.setattr(bonfire.bonfire, "get_apps_for_env", _mock_get_apps_for_env)
    elif source == FILE_SRC:
        monkeypatch.setattr(bonfire.bonfire, "get_appsfile_apps", lambda _: _target_apps())
    else:
        raise ValueError(f"invalid source '{source}' provided to mock function")


@pytest.mark.parametrize("local_config_method", ("merge", "override"))
@pytest.mark.parametrize("source", (APP_SRE_SRC, FILE_SRC))
def test_local_no_remote_target_apps_found(monkeypatch, source, local_config_method):
    local_cfg = {"apps": [val for _, val in _target_apps().items()]}
    _setup_monkeypatch(monkeypatch, source, local_cfg)
    actual = _get_apps_config(
        source=source,
        target_env="test_env_with_no_apps",
        ref_env=None,
        fallback_ref_env=None,
        local_config_path="na",
        local_config_method=local_config_method,
        preferred_params={},
    )
    assert actual == _target_apps()


@pytest.mark.parametrize("local_config_method", ("merge", "override"))
@pytest.mark.parametrize("source", (APP_SRE_SRC, FILE_SRC))
def test_new_local_app_added_to_remote_apps(monkeypatch, source, local_config_method):
    local_cfg = {
        "apps": [
            {
                "name": "appC",
                "components": [
                    {
                        "name": "appC_component1",
                        "host": "gitlab",
                        "repo": "someorg/somerepo",
                        "path": "template.yml",
                    }
                ],
            }
        ]
    }
    _setup_monkeypatch(monkeypatch, source, local_cfg)
    actual = _get_apps_config(
        source=source,
        target_env="test_target_env",
        ref_env=None,
        fallback_ref_env=None,
        local_config_path="na",
        local_config_method=local_config_method,
        preferred_params={},
    )
    expected = _target_apps()
    expected["appC"] = local_cfg["apps"][0]
    assert actual == expected


@pytest.mark.parametrize("source", (APP_SRE_SRC, FILE_SRC))
def test_new_local_component_merged(monkeypatch, source):
    local_cfg = {
        "apps": [
            {
                "name": "appB",
                "components": [
                    {
                        "name": "appB_component3",
                        "host": "gitlab",
                        "repo": "someorg/somerepo",
                        "path": "template.yml",
                    }
                ],
            }
        ]
    }
    _setup_monkeypatch(monkeypatch, source, local_cfg)
    actual = _get_apps_config(
        source=source,
        target_env="test_target_env",
        ref_env=None,
        fallback_ref_env=None,
        local_config_path="na",
        local_config_method="merge",
        preferred_params={},
    )
    expected = _target_apps()
    expected["appB"]["components"].append(local_cfg["apps"][0]["components"][0])
    assert actual == expected


@pytest.mark.parametrize("local_config_method", ("merge", "override"))
@pytest.mark.parametrize("source", (APP_SRE_SRC, FILE_SRC))
def test_empty_local_config(monkeypatch, source, local_config_method):
    local_cfg = {}
    _setup_monkeypatch(monkeypatch, source, local_cfg)
    actual = _get_apps_config(
        source=source,
        target_env="test_target_env",
        ref_env="test_ref_env",
        fallback_ref_env=None,
        local_config_path="na",
        local_config_method=local_config_method,
        preferred_params={},
    )
    assert actual == _target_apps_w_refs_subbed()


@pytest.mark.parametrize(
    "bad_local_cfg",
    (
        {"apps": [{"name": "bad_app", "components": {"oops": "bad_type"}}]},
        {"apps": "bad_type"},
        {"apps": [{"name": "bad_app", "components": [{"oops": "missing_keys"}]}]},
    ),
    ids=(
        "wrong_type_for_components_list",
        "wrong_type_for_apps_dict",
        "missing_keys_in_components_dict",
    ),
)
@pytest.mark.parametrize("local_config_method", ("merge", "override"))
@pytest.mark.parametrize("source", (APP_SRE_SRC, FILE_SRC))
def test_bad_local_config(monkeypatch, source, local_config_method, bad_local_cfg):
    _setup_monkeypatch(monkeypatch, source, bad_local_cfg)
    with pytest.raises(bonfire.utils.FatalError) as exc:
        _get_apps_config(
            source=source,
            target_env="test_env_with_no_apps",
            ref_env=None,
            fallback_ref_env=None,
            local_config_path="na",
            local_config_method=local_config_method,
            preferred_params={},
        )
        assert str(exc).startswith(bonfire.utils.SYNTAX_ERR)


@pytest.mark.parametrize("local_config_method", ("merge", "override"))
@pytest.mark.parametrize("source", (APP_SRE_SRC, FILE_SRC))
def test_master_branch_used_when_no_reference_app_found(monkeypatch, source, local_config_method):
    local_cfg = {}
    _setup_monkeypatch(monkeypatch, source, local_cfg)

    apps_config = _get_apps_config(
        source=source,
        target_env="test_target_env",
        ref_env="test_env_with_no_apps",
        fallback_ref_env=None,
        local_config_path="na",
        local_config_method=local_config_method,
        preferred_params={},
    )

    expected = _target_apps_w_refs_subbed()
    for _, app_config in expected.items():
        for component in app_config["components"]:
            component["ref"] = "master"

    assert apps_config == expected


@pytest.mark.parametrize("local_config_method", ("merge", "override"))
@pytest.mark.parametrize("source", (APP_SRE_SRC, FILE_SRC))
def test_fallback_reference_env(monkeypatch, source, local_config_method):
    local_cfg = {}
    _setup_monkeypatch(monkeypatch, source, local_cfg)

    apps_config = _get_apps_config(
        source=source,
        target_env="test_target_env",
        ref_env="test_env_with_no_apps",
        fallback_ref_env="test_ref_env",
        local_config_path="na",
        local_config_method=local_config_method,
        preferred_params={},
    )

    assert apps_config == _target_apps_w_refs_subbed()


@pytest.mark.parametrize("source", (APP_SRE_SRC, FILE_SRC))
def test_local_config_merge(monkeypatch, source):
    local_cfg = {
        "apps": [
            {
                "name": "appB",
                "components": [
                    {
                        "name": "appBcomponent2",
                        "parameters": {
                            "NEW_PARAM1": "NEW_VALUE1",
                            "NEW_PARAM2": "NEW_VALUE2",
                        },
                    },
                    # intentionally reversing order of components...
                    {
                        "name": "appBcomponent1",
                        "repo": "neworg/newrepo",
                        "ref": "a_new_ref",
                    },
                ],
            },
        ]
    }

    _setup_monkeypatch(monkeypatch, source, local_cfg)

    actual = _get_apps_config(
        source=source,
        target_env="test_target_env",
        ref_env=None,
        fallback_ref_env=None,
        local_config_path="na",
        local_config_method="merge",
        preferred_params={},
    )

    expected = _target_apps()
    for _, app_config in expected.items():
        for component in app_config["components"]:
            if component["name"] == "appBcomponent1":
                component["ref"] = "a_new_ref"
                component["repo"] = "neworg/newrepo"
            if component["name"] == "appBcomponent2":
                component["parameters"] = {
                    "EXISTING_PARAM1": "EXISTING_VALUE1",
                    "EXISTING_PARAM2": "EXISTING_VALUE2",
                    "NEW_PARAM1": "NEW_VALUE1",
                    "NEW_PARAM2": "NEW_VALUE2",
                }

    assert actual == expected


@pytest.mark.parametrize("source", (APP_SRE_SRC, FILE_SRC))
def test_local_config_override(monkeypatch, source):
    app_b = {
        "name": "appB",
        "components": [
            {
                "name": "appBcomponent2",
                "host": "gitlab",
                "repo": "somefork/appBcomponent2",
                "path": "deploy/new_template.yaml",
                "ref": "some_custom_ref",
                "parameters": {"SOME_PARAM": "SOME_VALUE"},
            },
        ],
    }
    local_cfg = {"apps": [app_b]}

    _setup_monkeypatch(monkeypatch, source, local_cfg)

    actual = _get_apps_config(
        source=source,
        target_env="test_target_env",
        ref_env=None,
        fallback_ref_env=None,
        local_config_path="na",
        local_config_method="override",
        preferred_params={},
    )

    expected = _target_apps()
    expected["appB"] = app_b

    assert actual == expected


@pytest.mark.parametrize("source", (APP_SRE_SRC, FILE_SRC))
def test_local_config_merge_update_param(monkeypatch, source):
    local_cfg = {
        "apps": [
            {
                "name": "appB",
                "components": [
                    {
                        "name": "appBcomponent2",
                        "parameters": {
                            "EXISTING_PARAM1": "NEW_VALUE1",
                            "NEW_PARAM2": "NEW_VALUE2",
                        },
                    },
                ],
            },
        ]
    }

    _setup_monkeypatch(monkeypatch, source, local_cfg)

    actual = _get_apps_config(
        source=source,
        target_env="test_target_env",
        ref_env=None,
        fallback_ref_env=None,
        local_config_path="na",
        local_config_method="merge",
        preferred_params={},
    )

    expected = _target_apps()
    component1_found = False
    for _, app_config in expected.items():
        for component in app_config["components"]:
            if component["name"] == "appBcomponent1":
                component1_found = True
            if component["name"] == "appBcomponent2":
                component["parameters"] = {
                    "EXISTING_PARAM1": "NEW_VALUE1",
                    "EXISTING_PARAM2": "EXISTING_VALUE2",
                    "NEW_PARAM2": "NEW_VALUE2",
                }

    assert component1_found  # ensure component 1 is not removed
    assert actual == expected
